from __future__ import annotations

from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import gdsfactory as gf


def plot_sparameters(
    df: DataFrame,
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
) -> None:
    """Plots Sparameters from a pandas DataFrame.

    Args:
        df: Sparameters pandas DataFrame.
        logscale: plots 20*log10(S).
        keys: list of keys to plot, plots all by default.

    .. plot::
        :include-source:

        import gdsfactory as gf
        import gdsfactory.simulation as sim

        df = sim.get_sparameters_data_lumerical(component=gf.components.mmi1x2)
        sim.plot.plot_sparameters(df, logscale=True)

    """
    w = df["wavelengths"] * 1e3
    keys = keys or [
        key for key in df.keys() if key.lower().startswith("s") and key.endswith("m")
    ]

    for key in keys:
        if key not in df:
            raise ValueError(f"{key} not in {df.keys()}")
        y = df[key]
        y = 20 * np.log10(y) if logscale else y
        plt.plot(w, y, label=key[:-1])
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("|S| (dB)") if logscale else plt.ylabel("|S|")
    plt.show()


def plot_imbalance2x2(df: DataFrame, port1: str = "s13m", port2: str = "s14m") -> None:
    """Plots imbalance in % for 2x2 coupler.

    Args:
        df: pandas DataFrame.
        port1: name.
        port2: name.

    """
    y1 = df[port1].values
    y2 = df[port2].values
    imbalance = y1 / y2
    x = df["wavelengths"] * 1e3
    plt.plot(x, 100 * abs(imbalance))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("imbalance (%)")
    plt.grid()


def plot_loss2x2(df: DataFrame, port1: str = "s13m", port2: str = "s14m") -> None:
    """Plots imbalance in % for 2x2 coupler.

    Args:
        df: pandas DataFrame.
        port1: name.
        port2: name.

    """
    y1 = df[port1].values
    y2 = df[port2].values
    x = df["wavelengths"] * 1e3
    plt.plot(x, abs(10 * np.log10(y1**2 + y2**2)))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("excess loss (dB)")


plot_loss1x2 = gf.partial(plot_loss2x2, port1="s13m", port2="o1@0,o2@0")
plot_imbalance1x2 = gf.partial(plot_imbalance2x2, port1="s13m", port2="s12m")


if __name__ == "__main__":
    import gdsfactory.simulation as sim

    df = sim.get_sparameters_data_lumerical(component=gf.components.mmi1x2)
    plot_sparameters(df, logscale=True)
    plt.show()



from __future__ import annotations

from simphony.libraries.sipann import Racetrack
from simphony.models import Subcircuit

from gdsfactory.simulation.simphony.plot_circuit import plot_circuit


def ring_single(
    wg_width: float = 0.5,
    gap: float = 0.2,
    length_x: float = 4,
    radius: float = 5,
    length_y: float = 2,
) -> Subcircuit:
    r"""Return single bus ring Model made of a ring coupler (cb: bottom).

    connected with:
    - 2 vertical straights (wl: left, wr: right)
    - 2 bend90 straights (bl: left, br: right)
    - 1 straight at the top (wt)

    .. code::

              wt (top)
              length_x
             /         \
            /           \
           |             |
           o1           o1 ___
                            |
          wl            wr  | length_y
                           _|_
           o2            o3
           |             |
            \           /
             \         /
           ---=========---
        o1 o1 length_x  o4 o2



    .. plot::
      :include-source:

      import gdsfactory as gf

      c = gf.components.ring_single(width=0.5, gap=0.2, length_x=4, radius=5, length_y=2)
      c.plot()


    .. plot::
        :include-source:

        import gdsfactory.simulation.simphony as gs
        import gdsfactory.simulation.simphony.components as gc

        c = gc.ring_single()
        gs.plot_circuit(c)
    """
    wg_width *= 1e-6
    gap *= 1e-6
    length_x *= 1e-6
    radius *= 1e-6
    length_y *= 1e-6

    racetrack = Racetrack(wg_width, 0.22e-6, radius, gap, length_x)

    racetrack.rename_pins("o1", "o2")
    return racetrack.circuit.to_subcircuit()


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    c = ring_single()
    plot_circuit(c)
    plt.show()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.typings import ComponentSpec, CrossSectionSpec

diagram = """

       | L0 |    L2        |

            ->-------------|
                           | pi * radius
       |-------------------|
       |
       |------------------->

       |        DL         |

"""


@gf.cell
def delay_snake(
    length: float = 1600.0,
    L0: float = 5.0,
    n: int = 2,
    bend: ComponentSpec = "bend_euler",
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    """Returns Snake with a starting straight and 90 bends.

    Input faces west output faces east.

    Args:
        length: delay length in um.
        L0: initial xoffset in um.
        n: number of loops.
        bend: bend spec.
        cross_section: cross_section spec.
        kwargs: cross_section settings.

    .. code::

       | L0 |    L2        |

            ->-------------|
                           | pi * radius
       |-------------------|
       |
       |------------------->

       |        DL         |
    """
    epsilon = 0.1
    bend90 = gf.get_component(bend, cross_section=cross_section, **kwargs)
    dy = bend90.info["dy"]
    DL = (length + L0 - n * (np.pi * dy + epsilon)) / (2 * n + 1)
    L2 = DL - L0
    if L2 < 0:
        raise ValueError(
            "Snake is too short: either reduce L0, increase "
            "the total length, or decrease n \n" + diagram
        )

    y = 0
    path = [(0, y), (L2, y)]
    for _i in range(n):
        y -= 2 * dy + epsilon
        path += [(L2, y), (-L0, y)]
        y -= 2 * dy + epsilon
        path += [(-L0, y), (L2, y)]

    path = [(round(_x, 3), round(_y, 3)) for _x, _y in path]

    c = gf.Component()
    route = round_corners(
        points=path, bend=bend90, cross_section=cross_section, **kwargs
    )
    c.add(route.references)
    c.add_port("o1", port=route.ports[0])
    c.add_port("o2", port=route.ports[1])
    return c


def test_delay_snake_length():
    length = 200.0
    c = delay_snake(n=1, length=length, cross_section="strip_no_pins")
    length_computed = c.area() / 0.5
    np.isclose(length, length_computed)
    return c


if __name__ == "__main__":
    c = test_delay_snake_length()
    # c = delay_snake(cross_section="strip_auto_widen", auto_widen_minimum_length=50)
    c.show(show_ports=True)



from __future__ import annotations

import pandas as pd

import gdsfactory as gf
from gdsfactory.typings import Component

debug = False
nlabels = 12


def mzi_te(**kwargs) -> Component:
    gc = gf.c.grating_coupler_elliptical_tm()
    c = gf.c.mzi_phase_shifter_top_heater_metal(delta_length=40)
    c = gf.routing.add_fiber_array(c, grating_coupler=gc, **kwargs)
    c = gf.routing.add_electrical_pads_shortest(c)
    return c


def test_grid() -> Component:
    c = gf.grid(
        [mzi_te()] * 2,
        decorator=gf.add_labels.add_labels_to_ports,
        add_ports_suffix=True,
        add_ports_prefix=False,
    )
    gdspath = c.write_gds()
    csvpath = gf.labels.write_labels.write_labels_gdstk(gdspath, debug=debug)

    df = pd.read_csv(csvpath)
    assert len(df) == nlabels, len(df)
    return c


def test_pack() -> Component:
    c = gf.pack(
        [mzi_te()] * 2,
        add_ports_suffix=True,
        add_ports_prefix=False,
    )[0]
    c = gf.add_labels.add_labels_to_ports(c)
    gdspath = c.write_gds()
    csvpath = gf.labels.write_labels.write_labels_gdstk(gdspath, debug=debug)
    df = pd.read_csv(csvpath)
    assert len(df) == nlabels, len(df)
    return c


if __name__ == "__main__":
    # c = test_grid()
    c = test_pack()
    c.show()
    # c = gf.pack(
    #     [mzi_te()] * 2,
    #     add_ports_suffix=True,
    #     add_ports_prefix=False,
    # )[0]
    # c = gf.add_labels.add_labels_to_ports(c)
    # gdspath = c.write_gds()
    # csvpath = gf.labels.write_labels.write_labels_gdstk(gdspath, debug=debug)
    # df = pd.read_csv(csvpath)



from __future__ import annotations

import meep as mp

from gdsfactory.config import logger
from gdsfactory.simulation import plot, port_symmetries
from gdsfactory.simulation.get_sparameters_path import get_sparameters_data_meep
from gdsfactory.simulation.gmeep.get_simulation import get_simulation
from gdsfactory.simulation.gmeep.meep_adjoint_optimization import (
    get_meep_adjoint_optimizer,
    run_meep_adjoint_optimizer,
)
from gdsfactory.simulation.gmeep.write_sparameters_grating import (
    write_sparameters_grating,
    write_sparameters_grating_batch,
    write_sparameters_grating_mpi,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep import (
    write_sparameters_meep,
    write_sparameters_meep_1x1,
    write_sparameters_meep_1x1_bend90,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep_batch import (
    write_sparameters_meep_batch,
    write_sparameters_meep_batch_1x1,
    write_sparameters_meep_batch_1x1_bend90,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep_mpi import (
    write_sparameters_meep_mpi,
    write_sparameters_meep_mpi_1x1,
    write_sparameters_meep_mpi_1x1_bend90,
)

logger.info(f"Meep {mp.__version__!r} installed at {mp.__path__!r}")

__all__ = [
    "get_meep_adjoint_optimizer",
    "get_simulation",
    "get_sparameters_data_meep",
    "run_meep_adjoint_optimizer",
    "write_sparameters_meep",
    "write_sparameters_meep_1x1",
    "write_sparameters_meep_1x1_bend90",
    "write_sparameters_meep_mpi",
    "write_sparameters_meep_mpi_1x1",
    "write_sparameters_meep_mpi_1x1_bend90",
    "write_sparameters_meep_batch",
    "write_sparameters_meep_batch_1x1",
    "write_sparameters_meep_batch_1x1_bend90",
    "write_sparameters_grating",
    "write_sparameters_grating_mpi",
    "write_sparameters_grating_batch",
    "plot",
    "port_symmetries",
]
__version__ = "0.0.3"



from typing import Optional

import numpy as np

from gdsfactory.pdk import get_layer_stack
from gdsfactory.technology import LayerStack
import gdsfactory as gf
import shapely
from shapely.affinity import translate
from shapely.ops import unary_union


class Parameter:
    def __init__(
        self,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
        nominal_value: Optional[float] = None,
        step: Optional[float] = None,
    ) -> None:
        """Generic parameter class for training Component models.

        Arguments:
            min_value: minimum value of the parameter.
            max_value: maximum value of the parameter.
            nominal_value: nominal value of the parameter.
            step: size of the step going from min_value to max_value when generating data.
        """

    def sample(self, rand_val: float = None):
        """Return a random value within a parameter allowable values.

        User can provide their own random number between 0 and 1 (mapping to a value between min and max), or default to uniform sampling.

        Arguments:
            rand_val: random value between 0 and 1, where 0 maps to min_value and 1 to max_value.

        """
        rand_val = rand_val or np.random.rand(1)[0]
        return self.min_value + (self.max_value - self.min_value) * rand_val

    def count(self):
        """Given min, max, and step, returns number of grid points."""
        return np.ceil(np.abs(self.max_value - self.min_value) / self.step)

    def arange(self):
        """Given min, max, and step, return array of values between min and max (inclusive)."""
        return np.arange(self.min_value, self.max_value + self.step / 2, self.step)

    def corners(self):
        """Returns an array of min, nominal, and max values of the parameter."""
        return np.array([self.min_value, self.nominal_value, self.max_value])


class LayerStackThickness(Parameter):
    def __init__(
        self,
        layerstack: Optional[LayerStack] = None,
        layername: Optional[str] = "core",
        **kwargs,
    ) -> None:
        """Layerstack thickness parameter.

        Arguments:
            layerstack: LayerStack
            layername: Name of the layer in the layerstack
            min_value: minimum value of the parameter. Default to layer thickness minus tolerance.
            max_value: maximum value of the parameter. Default to layer thickness plus tolerance.
            nominal_value: nominal value of the parameter. Default to layer thickness.
            step: size of the step going from min_value to max_value when generating data. Default to 3 steps between min and max.
        """
        super().__init__(**kwargs)
        self.layerstack = layerstack or get_layer_stack()
        self.layername = layername
        self.min_value = (
            kwargs["min_value"]
            or self.layerstack.layers[self.layername].thickness
            - self.layerstack.layers[self.layername].thickness_tolerance
        )
        self.max_value = (
            kwargs["max_value"]
            or self.layerstack.layers[self.layername].thickness
            + self.layerstack.layers[self.layername].thickness_tolerance
        )
        self.nominal_value = self.layerstack.layers[self.layername].thickness
        self.step = kwargs["step"] or np.abs(self.max_value - self.min_value) / 3
        self.current_value = None
        return None


class NamedParameter(Parameter):
    def __init__(self, **kwargs) -> None:
        """Parameter associated with the Component function signature or simulation (e.g. some physical dimension, wavelength)."""
        super().__init__(**kwargs)
        self.min_value = kwargs.get("min_value")
        self.max_value = kwargs.get("max_value")
        self.nominal_value = kwargs.get("nominal_value")
        self.step = kwargs.get("step")
        return None


class LithoParameter(Parameter):
    def __init__(
        self,
        type: str = "layer_dilation_erosion",
        layerstack: Optional[LayerStack] = None,
        layername: Optional[str] = "core",
        **kwargs,
    ) -> None:
        """Parameter associated with a morphological transformation of the Component.

        Can be parsed by the "perturb_geometry" method of the model builder to create a temporary new component just prior to simulation.

        Currently implemented transformations:
            * Erosion and dilation (type = "layer_dilation_erosion")
            * Layer translation offset (type = "layer_x_offset" and "layer_y_offset")
            * Corner rounding (type = "layer_round_corners")
        """
        self.min_value = kwargs.get("min_value")
        self.max_value = kwargs.get("max_value")
        self.nominal_value = kwargs.get("nominal_value")
        self.step = kwargs.get("step")
        layerstack = layerstack or get_layer_stack()

        self.layer = layerstack[layername].layer
        self.type = type

        return None

    def get_transformation(self, component, value):
        if self.type == "layer_dilation_erosion":
            return self.layer_dilation_erosion(component, value)
        elif self.type == "layer_x_offset":
            return self.layer_x_offset(component, value)
        elif self.type == "layer_y_offset":
            return self.layer_y_offset(component, value)
        elif self.type == "layer_round_corners":
            return self.layer_round_corners(component, value)
        else:
            raise ValueError(
                'LithoParameter requires type = "layer_dilation_erosion", "layer_x_offset", "layer_y_offset", or "layer_round_corners"'
            )

    def layer_dilation_erosion(self, component, dilation_value):
        temp_component = gf.Component()
        for layer, layer_polygons in component.get_polygons(by_spec=True).items():
            if layer == self.layer:
                # Make sure all layer polygons are fused properly
                shapely_polygons = [
                    shapely.geometry.Polygon(polygon) for polygon in layer_polygons
                ]
                shapely_polygons = unary_union(shapely_polygons)
                # Apply transformation
                for shapely_polygon in (
                    shapely_polygons.geoms
                    if hasattr(shapely_polygons, "geoms")
                    else [shapely_polygons]
                ):
                    buffered_polygons = shapely_polygon.buffer(
                        dilation_value, join_style=2
                    )
                    for buffered_polygon in (
                        buffered_polygons.geoms
                        if hasattr(buffered_polygons, "geoms")
                        else [buffered_polygons]
                    ):
                        temp_component.add_polygon(
                            buffered_polygon.exterior.coords, layer=layer
                        )
            else:
                for layer_polygon in layer_polygons:
                    temp_component.add_polygon(layer_polygon, layer=layer)
        # Transform ports
        ports = []
        for port in component.get_ports():
            if port.layer == self.layer:
                port.width += 2 * dilation_value
                old_center_x, old_center_y = port.center
                new_center_x = (
                    old_center_x - np.cos(np.radians(port.orientation)) * dilation_value
                )
                new_center_y = (
                    old_center_y - np.sin(np.radians(port.orientation)) * dilation_value
                )
                port.center = [new_center_x, new_center_y]
            ports.append(port)
        temp_component.add_ports(ports=ports)
        return temp_component

    def layer_x_offset(self, component, offset_value):
        temp_component = gf.Component()
        for layer, layer_polygons in component.get_polygons(by_spec=True).items():
            for layer_polygon in layer_polygons:
                if layer == self.layer:
                    shapely_polygon = shapely.geometry.Polygon(layer_polygon)
                    translated_polygon = translate(
                        shapely_polygon, xoff=offset_value, yoff=0.0
                    )
                    temp_component.add_polygon(
                        translated_polygon.exterior.coords, layer=layer
                    )
                else:
                    temp_component.add_polygon(layer_polygon, layer=layer)
        # Transform ports
        ports = []
        for port in component.get_ports():
            if port.layer == self.layer:
                old_center_x, old_center_y = port.center
                new_center_x = old_center_x + offset_value
                port.center = [new_center_x, old_center_y]
            ports.append(port)
        temp_component.add_ports(ports=ports)
        return temp_component

    def layer_y_offset(self, component, offset_value):
        temp_component = gf.Component()
        for layer, layer_polygons in component.get_polygons(by_spec=True).items():
            for layer_polygon in layer_polygons:
                if layer == self.layer:
                    shapely_polygon = shapely.geometry.Polygon(layer_polygon)
                    translated_polygon = translate(
                        shapely_polygon, xoff=0.0, yoff=offset_value
                    )
                    temp_component.add_polygon(
                        translated_polygon.exterior.coords, layer=layer
                    )
                else:
                    temp_component.add_polygon(layer_polygon, layer=layer)
        # Transform ports
        ports = []
        for port in component.get_ports():
            if port.layer == self.layer:
                old_center_x, old_center_y = port.center
                new_center_y = old_center_y + offset_value
                port.center = [old_center_x, new_center_y]
            ports.append(port)
        temp_component.add_ports(ports=ports)
        return temp_component

    def layer_round_corners(self, component, round_value):
        temp_component = gf.Component()
        for layer, layer_polygons in component.get_polygons(by_spec=True).items():
            if layer == self.layer:
                # Make sure all layer polygons are fused properly
                shapely_polygons = [
                    shapely.geometry.Polygon(polygon) for polygon in layer_polygons
                ]
                shapely_polygons = unary_union(shapely_polygons)
                # Apply transformation
                for shapely_polygon in (
                    shapely_polygons.geoms
                    if hasattr(shapely_polygons, "geoms")
                    else [shapely_polygons]
                ):
                    buffered_polygon = (
                        shapely_polygon.buffer(round_value, join_style=1)
                        .buffer(-2 * round_value, join_style=1)
                        .buffer(round_value, join_style=1)
                    )
                    temp_component.add_polygon(
                        buffered_polygon.exterior.coords, layer=layer
                    )
            else:
                for layer_polygon in layer_polygons:
                    temp_component.add_polygon(layer_polygon, layer=layer)
        # Transform ports
        ports = []
        for port in component.get_ports():
            if port.layer == self.layer:
                # Patch port
                patch_polygon_x1, patch_polygon_y1 = port.center
                patch_polygon_x2 = (
                    patch_polygon_x1
                    + np.cos(np.radians(port.orientation)) * round_value
                )
                patch_polygon_y2 = (
                    patch_polygon_y1
                    + np.sin(np.radians(port.orientation)) * round_value
                )
                P = gf.Path(
                    [
                        [patch_polygon_x1, patch_polygon_y1],
                        [patch_polygon_x2, patch_polygon_y2],
                    ],
                )
                temp_component << gf.path.extrude(P, width=port.width, layer=port.layer)
            ports.append(port)
        temp_component.add_ports(ports=ports)
        return temp_component


if __name__ == "__main__":
    c = gf.Component("myComponent")
    poly1a = c.add_polygon(
        [[2.8, 3], [5, 3], [5, 0.8]],
        layer=1,
    )
    poly1b = c.add_polygon(
        [
            [2, 0],
            [2, 2],
            [4, 2],
            [4, 0],
        ],
        layer=1,
    )
    poly1c = c.add_polygon(
        [
            [0, 0.5],
            [0, 1.5],
            [3, 1.5],
            [3, 0.5],
        ],
        layer=1,
    )
    poly2 = c.add_polygon(
        [
            [0, 0],
            [5, 0],
            [5, 3],
            [0, 3],
        ],
        layer=2,
    )
    poly3 = c.add_polygon(
        [
            [2.5, -2],
            [3.5, -2],
            [3.5, -0.1],
            [2.5, -0.1],
        ],
        layer=1,
    )
    c.add_port(name="o1", center=(0, 1), width=1, orientation=0, layer=1)
    c.add_port(name="o2", center=(3, -2), width=1, orientation=90, layer=1)
    c.show(show_ports=True)

    param = LithoParameter(layername="core")
    eroded_c = param.layer_dilation_erosion(c, 0.2)
    eroded_c.show(show_ports=True)

    param = LithoParameter(layername="core")
    eroded_c = param.layer_dilation_erosion(c, -0.2)
    eroded_c.show(show_ports=True)

    param = LithoParameter(layername="core")
    eroded_c = param.layer_x_offset(c, 0.5)
    eroded_c.show(show_ports=True)

    param = LithoParameter(layername="core")
    eroded_c = param.layer_y_offset(c, 0.5)
    eroded_c.show(show_ports=True)

    param = LithoParameter(layername="core")
    eroded_c = param.layer_round_corners(c, 0.2)
    eroded_c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component


@cell
def logo(text: str = "GDSFACTORY") -> Component:
    """Returns GDSfactory logo."""
    c = Component()
    elements = []
    for i, letter in enumerate(text):
        c << gf.components.text(letter, layer=(i + 1, 0), size=10)
        elements.append(c)

    c.distribute(
        elements="all",  # either 'all' or a list of objects
        direction="x",  # 'x' or 'y'
        spacing=1,
        separation=True,
    )
    return c


if __name__ == "__main__":
    c = logo()
    c.show(show_ports=True)



"""
Generates a ring resonator based on a series of given cross sections.

This is useful to generate interleaved junction rings or rings with relatively
complex junction profiles
"""

from functools import partial
from typing import Dict, Optional

import numpy as np
import gdsfactory as gf
from gdsfactory.typings import CrossSectionSpec, Union, List, Tuple, Floats
from gdsfactory.components.bend_circular import bend_circular
from gdsfactory.components.straight import straight
from gdsfactory.cell import cell_without_validator

def_dict = {"A": "rib", "B": "strip"}
def_ang_dict = {"A": 6, "B": 6}


@cell_without_validator
def ring_section_based(
    gap: Union[float, Floats] = 0.3,
    radius: float = 5.0,
    add_drop: bool = False,
    cross_sections: Dict[str, CrossSectionSpec] = def_dict,
    cross_sections_sequence: Union[str, List[str], Tuple[str, ...]] = "AB",
    cross_sections_angles: Optional[Dict[str, float]] = def_ang_dict,
    start_cross_section: Optional[CrossSectionSpec] = None,
    start_angle: Optional[float] = 10.0,
    drop_cross_section: Optional[CrossSectionSpec] = None,
    bus_cross_section: CrossSectionSpec = "strip",
    ang_res: Optional[int] = 0.1,
) -> gf.Component:
    """Returns a ring made of the specified cross sections.

    We start with start_cross section if indicated, then repeat the sequence in
    cross_section_sequence until the whole ring is filled.

    Args:
        gap: bus waveguide - ring gap.
        radius: ring radius.
        add_drop: if True, we draw an add-drop ring
        cross_sections: dictionary of cross sections to add consecutively
            to the ring until the ring is filled. Keys should be single character.
        cross_sections_sequence: sequence to follow filling the ring.
            Ex: "AB" means we will put first section A, then section B,
            then section A again... until the ring is filled
            If we have a drop, then this can be a list of two strings,
            where the first sequence is for one side of the ring
            and the second for the other side of the ring
        cross_sections_angles: angular extent of each cross section in the
            cross_sections dictionary (deg). If not indicated, then we assume that
            the sequence is only repeated once and calculate the necessary angular
            length
        start_cross_section: it is likely that the cross section at the ring-bus junction
            is different than the sequence we want to repeat. If that's the case, then
            here you indicate the initial cross section.
        start_angle: angular extent of the initial cross section (deg)
        drop_cross_section: cross section for the drop port. If not indicated, we assume
            it is the same as init_cross_section.
        bus_cross_section: cross section for the bus waveguide.
        ang_res: angular resolution to draw the bends for each section
    """

    c = gf.Component()

    # First of all we need to do a bunch of checks
    angular_extent_sequence = 360

    if not isinstance(gap, Union[List, Tuple]):
        gap = [gap] * 2

    # See if we need to add initial cross sections
    if start_cross_section is not None:
        start_xs = gf.get_cross_section(start_cross_section)
        angular_extent_sequence -= start_angle

    if add_drop:
        if start_cross_section is not None or drop_cross_section is not None:
            angular_extent_sequence -= start_angle

        if not isinstance(cross_sections_sequence, Union[List, Tuple]):
            cross_sections_sequence = [cross_sections_sequence] * 2

    if cross_sections_angles is None:
        if add_drop:
            n_sections_0 = len(cross_sections_sequence[0])
            sing_sec_ang_1 = angular_extent_sequence / (2 * n_sections_0)
            n_sections_1 = len(cross_sections_sequence[1])
            sing_sec_ang_2 = angular_extent_sequence / (2 * n_sections_1)
            cross_sections_angles = {
                elem: sing_sec_ang_1 for elem in cross_sections_sequence[0]
            }
            for elem in cross_sections_sequence[1]:
                cross_sections_angles[elem] = sing_sec_ang_2

        else:
            n_sections = len(cross_sections_sequence)
            sing_sec_ang = angular_extent_sequence / n_sections
            cross_sections_angles = {
                elem: sing_sec_ang for elem in cross_sections_sequence
            }

    # Now make sure that the specified angular extents of the sections
    # are compatible with the ring extent (360 degree)
    if not add_drop:
        sing_seq_angular_extent = np.sum(
            [cross_sections_angles[sec] for sec in cross_sections_sequence]
        )
        if not (angular_extent_sequence / sing_seq_angular_extent).is_integer():
            raise ValueError(
                "The specified sequence angles do not result in an integer number "
                "of sequences fitting in the ring."
            )
        else:
            n_repeats_seq = int(angular_extent_sequence / sing_seq_angular_extent)

    else:
        sing_seq_angular_extent_0 = np.sum(
            [cross_sections_angles[sec] for sec in cross_sections_sequence[0]]
        )

        sing_seq_angular_extent_1 = np.sum(
            [cross_sections_angles[sec] for sec in cross_sections_sequence[1]]
        )

        if np.round(
            angular_extent_sequence / (sing_seq_angular_extent_0 * 2), 5
        ).is_integer():
            n_repeats_seq_0 = int(
                angular_extent_sequence / (sing_seq_angular_extent_0 * 2)
            )
        else:
            raise ValueError(
                "The specified sequence 1 angles do not result in an integer "
                "number of sequences fitting in the ring."
            )

        if np.round(
            angular_extent_sequence / (sing_seq_angular_extent_1 * 2)
        ).is_integer():
            n_repeats_seq_1 = int(
                angular_extent_sequence / (sing_seq_angular_extent_1 * 2)
            )
        else:
            raise ValueError(
                "The specified sequence 2 angles do not result in an integer "
                "number of sequences fitting in the ring."
            )

    # Now we are ready to construct the ring

    # We need to create a circular bend for each section
    sections_dict = {}

    for key, xsec in cross_sections.items():
        ang = cross_sections_angles[key]
        b = bend_circular(
            angle=ang,
            with_bbox=False,
            cross_section=xsec,
            radius=radius,
            npoints=np.round(ang / ang_res) + 1 if ang_res is not None else None,
        )

        sections_dict[key] = (b, "o1", "o2")

    if start_cross_section is not None:
        b = bend_circular(
            angle=start_angle,
            with_bbox=False,
            cross_section=start_xs,
            radius=radius,
            npoints=np.round(ang / ang_res) + 1 if ang_res is not None else None,
        )
        if "0" in sections_dict:
            raise ValueError(
                "Please do not have '0' as a key for the cross_sections dict"
            )
        sections_dict["0"] = (b, "o1", "o2")

    if drop_cross_section is not None:
        b = bend_circular(
            angle=start_angle,
            with_bbox=False,
            cross_section=gf.get_cross_section(drop_cross_section),
            radius=radius,
            npoints=np.round(ang / ang_res) + 1 if ang_res is not None else None,
        )
        if "1" in sections_dict:
            raise ValueError(
                "Please do not have '1' as a key for the cross_sections dict"
            )
        sections_dict["1"] = (b, "o1", "o2")

    # Now we just need to generate a chain of characters
    # to creae the sequence

    sequence = ""

    if start_cross_section is not None:
        sequence += "0"

    if not add_drop:
        sequence += cross_sections_sequence * n_repeats_seq

    else:
        sequence += cross_sections_sequence[0] * n_repeats_seq_0

        if drop_cross_section is None and start_cross_section is not None:
            sequence += "0"
        elif drop_cross_section is not None:
            sequence += "1"

        seq_2 = cross_sections_sequence[1] * n_repeats_seq_1
        sequence += seq_2

    ring = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=sections_dict
    )

    r = c << ring

    # Rotate so that the first section is centered at the add bus
    if start_cross_section is not None:
        ring = ring.rotate(
            -start_angle / 2
        )  # also change the component for later bound extraction
        r.rotate(-start_angle / 2)
        ring_center = [
            -radius * np.sin(np.radians(-start_angle / 2)),
            radius * np.cos(np.radians(-start_angle / 2)),
        ]
    else:
        ring = ring.rotate(-cross_sections_angles[sequence[0]] / 2)
        r.rotate(-cross_sections_angles[sequence[0]] / 2)
        ring_center = [
            -radius * np.sin(np.radians(-cross_sections_angles[sequence[0]] / 2)),
            radius * np.cos(np.radians(-cross_sections_angles[sequence[0]] / 2)),
        ]
    c.info["ring_center"] = gf.snap.snap_to_grid(ring_center)
    c.info["ring_radius"] = radius

    # Add bus waveguides

    # Figure out main waveguiding layer of the ring at the ring-bus interface
    input_xs_layer = (
        gf.get_cross_section(start_cross_section).layer
        if start_cross_section
        else gf.get_cross_section(
            cross_sections[cross_sections_sequence[0][0]]
            if add_drop
            else cross_sections[cross_sections_sequence[0]]
        ).layer
    )
    ring_guide_add = ring.extract([input_xs_layer])

    if drop_cross_section:
        drop_xs_layer = gf.get_cross_section(drop_cross_section).layer
    elif start_cross_section:
        drop_xs_layer = input_xs_layer
    else:
        drop_xs_layer = gf.get_cross_section(
            cross_sections[cross_sections_sequence[1][0]]
        ).layer

    ring_guide_add = ring.extract([input_xs_layer])
    ring_guide_drop = ring.extract([drop_xs_layer])

    # Add bus waveguides
    s = straight(length=ring.xsize, cross_section=bus_cross_section)

    # Figure out main waveguiding layer of the bus at the ring-bus interface
    input_xs_width = gf.get_cross_section(bus_cross_section).width

    s_add = c << s
    s_add.x = r.x
    s_add.ymax = ring_guide_add.ymin - gap[0] + s.ysize / 2 - input_xs_width / 2

    if add_drop:
        # s.mirror((0, 1))
        s_drop = c << s
        s_drop.mirror_y()
        s_drop.x = r.x
        s_drop.ymin = ring_guide_drop.ymax + gap[1] - s.ysize / 2 + input_xs_width / 2

    # Add ports
    c.add_port("o1", port=s_add.ports["o1"], orientation=180)
    c.add_port("o2", port=s_add.ports["o2"], orientation=0)
    if add_drop:
        c.add_port("o3", port=s_drop.ports["o1"], orientation=180)
        c.add_port("o4", port=s_drop.ports["o2"], orientation=0)

    return c


if __name__ == "__main__":
    from gdsfactory.cross_section import rib

    # rib = c << straight(length = 100, cross_section="rib")
    # b_rib = c << bend_circular(angle = 10,
    #                       with_bbox = True,
    #                       cross_section = "slot",
    #                       radius=5)
    # strip = c << straight(length = 100, cross_section="strip")
    # strip.ymax = b_rib.ymin - 10

    c = ring_section_based(
        gap=0.3,
        radius=5.0,
        add_drop=True,
        cross_sections={"A": "rib", "B": "slot"},
        cross_sections_sequence=["AB", "BA"],
        cross_sections_angles={"A": 17, "B": 17},
        start_cross_section=partial(rib, width=0.65),
        start_angle=10.0,
        drop_cross_section=partial(rib, width=0.55),
        bus_cross_section="strip",
    )

    # b = c << bend_circular(angle = 15,
    #                     with_bbox = True,
    #                     cross_section = "strip",)
    # print(b.ports)
    # b.rotate(-7.5)
    # c = ring_section_based(add_drop=True)

    c.show(show_ports=True)
    # print(c.info["ring_center"])



from __future__ import annotations

from typing import Callable, List, Optional, Tuple, Union

import gdstk
import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component, ComponentReference
from gdsfactory.component_layout import _rotate_points
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.cross_section import strip
from gdsfactory.port import select_ports_optical
from gdsfactory.routing.route_fiber_array import route_fiber_array
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, Label


def route_fiber_single(
    component: Component,
    fiber_spacing: float = 50.0,
    grating_coupler: ComponentSpec = grating_coupler_te,
    min_input_to_output_spacing: float = 200.0,
    optical_routing_type: int = 1,
    port_names: Optional[Tuple[str, ...]] = None,
    excluded_ports: Optional[Tuple[str, ...]] = None,
    component_name: Optional[str] = None,
    select_ports: Callable = select_ports_optical,
    cross_section: CrossSectionSpec = strip,
    **kwargs,
) -> Tuple[List[Union[ComponentReference, Label]], List[ComponentReference]]:
    """Returns route Tuple(references, grating couplers) for single fiber input/output.

    Args:
        component: to add grating couplers.
        fiber_spacing: between grating couplers.
        grating_coupler: grating coupler Spec
        min_input_to_output_spacing: so opposite fibers do not touch
        optical_routing_type: 0 (basic), 1 (standard), 2 (looks at ports)
        port_names: port labels that need connection
        excluded_ports: ports excluded from routing
        component_name: Optional component name.
        select_ports: function to select ports.
        cross_section: spec.
        kwargs: cross_section settings

    Returns:
        elements: list of ComponentReferences for routes and labels
        grating_couplers: list of grating_couplers references


    .. code::
              _________
             |         |_E1
          W0_|         |
             |         |_E0
             |_________|


         rotates +90 deg and routes West ports to South

         the rest of the original ports (East, North, South) will route south

         it calls route_fiber_array twice

         route_fiber_array is designed to route ports south

               E1  E0
              _|___|_
             |       |
             |       |
             |       |
             |       |
             |       |
             |       |
             |_______|
                 |
                 W0     1st part routes West ports south

        then rotates 180 and routes the rest of the ports North

    """
    if not select_ports(component.ports):
        raise ValueError(f"No ports for {component.name}")

    component = component.copy()
    component_copy = component.copy()

    if port_names is None:
        optical_ports = select_ports(component.ports)
    else:
        optical_ports = [component.ports[lbl] for lbl in port_names]

    excluded_ports = excluded_ports or []
    optical_ports = {
        p.name: p for p in optical_ports.values() if p.name not in excluded_ports
    }
    if isinstance(grating_coupler, list):
        grating_couplers = [gf.call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = gf.call_if_func(grating_coupler)
        N = len(optical_ports)

        grating_couplers = [grating_coupler] * N

    gc_port2center = getattr(grating_coupler, "port2center", grating_coupler.xsize / 2)
    if component.xsize + 2 * gc_port2center < min_input_to_output_spacing:
        fanout_length = (
            gf.snap.snap_to_grid(
                min_input_to_output_spacing - component.xsize - 2 * gc_port2center, 10
            )
            / 2
        )
    else:
        fanout_length = None

    # route WEST ports to south
    component_west_ports = Component()
    ref = component_west_ports << component
    ref.rotate(90)
    south_ports = ref.get_ports_dict(orientation=270)
    component_west_ports.ports = south_ports

    if len(south_ports):
        (
            elements_south,
            gratings_south,
            ports_grating_input_waveguide_south,
            ports_loopback_south,
            ports_component_south,
        ) = route_fiber_array(
            component=component_west_ports,
            with_loopback=False,
            fiber_spacing=fiber_spacing,
            fanout_length=fanout_length,
            grating_coupler=grating_couplers[0],
            optical_routing_type=optical_routing_type,
            component_name=component_name,
            cross_section=cross_section,
            select_ports=select_ports,
            **kwargs,
        )

    # route non WEST ports north
    component = gf.Component()
    component_ref = component << component_copy
    component_ref.rotate(-90)
    component.add_ports(component_ref.ports)
    for port_already_routed in south_ports.values():
        component.ports.pop(port_already_routed.name)

    component.ports = select_ports(component.ports)

    (
        elements_north,
        gratings_north,
        ports_grating_input_waveguide_north,
        ports_loopback_north,
        ports_component_north,
    ) = route_fiber_array(
        component=component,
        with_loopback=False,
        fiber_spacing=fiber_spacing,
        fanout_length=fanout_length,
        grating_coupler=grating_couplers[1:],
        optical_routing_type=optical_routing_type,
        component_name=component_name,
        cross_section=cross_section,
        select_ports=select_ports,
        **kwargs,
    )
    for e in elements_north:
        if isinstance(e, list):
            for ei in e:
                if isinstance(ei, gdstk.Label):
                    ei.rotation = np.mod(ei.rotation + np.pi, 2 * np.pi)
                    print(ei.rotation)
                    ei.origin = _rotate_points(
                        ei.origin,
                        angle=np.rad2deg(ei.rotation),
                    )

                    elements_south.append(ei)
                else:
                    elements_south.append(ei.rotate(180))
        elif isinstance(e, gdstk.Label):
            ei = e
            ei.rotation = np.mod(ei.rotation + np.pi, 2 * np.pi)
            ei.origin = _rotate_points(
                ei.origin,
                angle=np.rad2deg(ei.rotation),
            )
            elements_south.append(e)
        else:
            elements_south.append(e.rotate(180))

    if len(gratings_north) > 0:
        for io in gratings_north[0]:
            gratings_south.append(io.rotate(180))

    ports_grating_input_waveguide = (
        ports_grating_input_waveguide_north + ports_grating_input_waveguide_south
    )
    ports_component = ports_component_north + ports_component_south
    return (
        elements_south,
        gratings_south,
        ports_grating_input_waveguide,
        ports_component,
    )


if __name__ == "__main__":
    gcte = gf.components.grating_coupler_te
    gctm = gf.components.grating_coupler_tm

    c = gf.components.cross(length=500)
    c = gf.components.ring_double()
    c = gf.components.crossing()
    c = gf.components.rectangle()

    # elements, gc = route_fiber_single(
    #     c, grating_coupler=[gcte, gctm, gcte, gctm],
    # )

    layer = (2, 0)
    c = gf.components.straight(width=2, length=500)
    c = gf.components.mmi1x2(length_mmi=167)
    c = gf.components.ring_single(length_x=167)
    c = gf.components.mmi2x2()
    # c = gf.components.spiral_inner_io_fiber_single()

    gc = gf.components.grating_coupler_elliptical_te(layer=layer)
    (
        elements,
        gc,
        ports_grating_input_waveguide,
        ports_component,
    ) = route_fiber_single(
        c,
        # grating_coupler=[gc, gc, gc, gc],
        radius=10,
        layer=layer,
    )

    cc = gf.Component("sample_route_fiber_single")
    cr = cc << c
    cr.rotate(90)

    for e in elements:
        cc.add(e)
    for e in gc:
        cc.add(e)

    cc.add_ports(ports_grating_input_waveguide)
    cc.show(show_ports=True)

    # layer = (31, 0)
    # c = gf.components.mmi2x2()
    # c = gf.components.straight(width=2, length=500)
    # gc = gf.components.grating_coupler_elliptical_te(layer=layer)
    # elements, gc = route_fiber_single(
    #     c,
    #     grating_coupler=[gc, gc, gc, gc],
    #     radius=10,
    #     layer=layer,
    # )
    # cc = gf.Component("sample_route_fiber_single")
    # cr = cc << c.rotate(90)
    # for e in elements:
    #     cc.add(e)
    # for e in gc:
    #     cc.add(e)
    # cc.show(show_ports=True)



"""Register materials."""
from __future__ import annotations

from typing import Callable, Dict, Tuple, Union

import numpy as np

MaterialSpec = Union[str, float, Tuple[float, float], Callable]

material_name_to_meep: Dict[str, MaterialSpec] = {
    "si": "Si",
    "sin": "Si3N4_NIR",
    "sio2": "SiO2",
}

material_name_to_lumerical: Dict[str, MaterialSpec] = {
    "si": "Si (Silicon) - Palik",
    "sio2": "SiO2 (Glass) - Palik",
    "sin": "Si3N4 (Silicon Nitride) - Phillip",
}


# default materials
def si(wav: np.ndarray) -> np.ndarray:
    """Silicon crystalline."""
    from gdsfactory.simulation.gtidy3d.materials import si

    return si(wav)


def sio2(wav: np.ndarray) -> np.ndarray:
    """Silicon oxide."""
    from gdsfactory.simulation.gtidy3d.materials import sio2

    return sio2(wav)


def sin(wav: np.ndarray) -> np.ndarray:
    """Silicon Nitride."""
    from gdsfactory.simulation.gtidy3d.materials import sin

    return sin(wav)


materials_index = {"si": si, "sio2": sio2, "sin": sin}

__all__ = [
    "material_name_to_meep",
    "material_name_to_lumerical",
    "materials_index",
    "si",
    "sio2",
    "sin",
]



from typing import Any, Dict, List, Optional, Tuple, Callable, Union
from collections import defaultdict

from typing_extensions import Literal
from pydantic import BaseModel, Field

from gdsfactory.technology.layer_views import LayerViews

MaterialSpec = Union[str, float, Tuple[float, float], Callable]


class LayerLevel(BaseModel):
    """Level for 3D LayerStack.

    Parameters:
        layer: (GDSII Layer number, GDSII datatype).
        thickness: layer thickness in um.
        thickness_tolerance: layer thickness tolerance in um.
        zmin: height position where material starts in um.
        material: material name.
        sidewall_angle: in degrees with respect to normal.
        sidewall_angle_tolerance: in degrees.
        width_to_z: if sidewall_angle, relative z-position
            (0 --> zmin, 1 --> zmin + thickness).
        z_to_bias: parametrizes shrinking/expansion of the design GDS layer
            when extruding from zmin (0) to zmin + thickness (1).
            Defaults no buffering [[0, 1], [0, 0]].
        mesh_order: lower mesh order (1) will have priority over higher
            mesh order (2) in the regions where materials overlap.
        refractive_index: refractive_index
            can be int, complex or function that depends on wavelength (um).
        type: grow, etch, implant, or background.
        mode: octagon, taper, round.
            https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
        into: etch into another layer.
            https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
        doping_concentration: for implants.
        resistivity: for metals.
        bias: in um for the etch. Can be a single number or 2 numbers (bias_x, bias_y)
        derived_layer: Optional derived layer, used for layer_type='etch' to define the slab.
        info: simulation_info and other types of metadata.
    """

    layer: Optional[Tuple[int, int]]
    thickness: float
    thickness_tolerance: Optional[float] = None
    zmin: float
    material: Optional[str] = None
    sidewall_angle: float = 0.0
    sidewall_angle_tolerance: Optional[float] = None
    width_to_z: float = 0.0
    z_to_bias: Optional[Tuple[List[float], List[float]]] = None
    mesh_order: int = 3
    layer_type: Literal["grow", "etch", "implant", "background"] = "grow"
    mode: Optional[Literal["octagon", "taper", "round"]] = None
    into: Optional[List[str]] = None
    doping_concentration: Optional[float] = None
    resistivity: Optional[float] = None
    bias: Optional[Union[Tuple[float, float], float]] = None
    derived_layer: Optional[Tuple[int, int]] = None
    info: Dict[str, Any] = {}


class LayerStack(BaseModel):
    """For simulation and 3D rendering.

    Parameters:
        layers: dict of layer_levels.
    """

    layers: Optional[Dict[str, LayerLevel]] = Field(default_factory=dict)

    def __init__(self, **data: Any):
        """Add LayerLevels automatically for subclassed LayerStacks."""
        super().__init__(**data)

        for field in self.dict():
            val = getattr(self, field)
            if isinstance(val, LayerLevel):
                self.layers[field] = val

    def get_layer_to_thickness(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to thickness (um)."""
        layer_to_thickness = {}

        for level in self.layers.values():
            layer = level.layer

            if layer and level.thickness:
                layer_to_thickness[layer] = level.thickness
            elif hasattr(level, "operator"):
                layer_to_thickness[level.layer] = level.thickness

        return layer_to_thickness

    def get_component_with_derived_layers(self, component):
        """Returns component with derived layers."""
        import gdstk

        unetched_layers = [
            layer_name
            for layer_name, level in self.layers.items()
            if level.layer and level.layer_type == "grow"
        ]
        etch_layers = [
            layer_name
            for layer_name, level in self.layers.items()
            if level.layer and level.layer_type == "etch"
        ]

        # remove all etched layers from the grown layers
        unetched_layers_dict = defaultdict(list)
        for layer_name in etch_layers:
            level = self.layers[layer_name]
            into = level.into or []
            for layer_name_etched in into:
                unetched_layers_dict[layer_name_etched].append(layer_name)
                if layer_name_etched in unetched_layers:
                    unetched_layers.remove(layer_name_etched)

        component_layers = component.get_layers()

        # Define pure grown layers
        unetched_layer_numbers = [
            self.layers[layer_name].layer
            for layer_name in unetched_layers
            if self.layers[layer_name].layer in component_layers
        ]
        component_derived = component.extract(unetched_layer_numbers)

        # Define unetched layers
        polygons_to_remove = []
        for unetched_layer_name, unetched_layers in unetched_layers_dict.items():
            layer = self.layers[unetched_layer_name].layer
            polygons = component.get_polygons(by_spec=layer)

            # Add all the etching layers (OR)
            for etching_layers in unetched_layers:
                layer = self.layers[etching_layers].layer
                B_polys = component.get_polygons(by_spec=layer)
                polygons_to_remove = gdstk.boolean(
                    operand1=polygons_to_remove,
                    operand2=B_polys,
                    operation="or",
                    layer=layer[0],
                    datatype=layer[1],
                )

                derived_layer = self.layers[etching_layers].derived_layer
                if derived_layer:
                    slab_polygons = gdstk.boolean(
                        operand1=polygons,
                        operand2=B_polys,
                        operation="and",
                        layer=derived_layer[0],
                        datatype=derived_layer[1],
                    )
                    component_derived.add(slab_polygons)

            # Remove all etching layers
            layer = self.layers[unetched_layer_name].layer
            polygons = component.get_polygons(by_spec=layer)
            unetched_polys = gdstk.boolean(
                operand1=polygons,
                operand2=polygons_to_remove,
                operation="not",
                layer=layer[0],
                datatype=layer[1],
            )
            component_derived.add(unetched_polys)

        component_derived.add_ports(component.ports)
        component_derived.name = f"{component.name}_derived_layers"
        return component_derived

    def get_layer_to_zmin(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to z min position (um)."""
        return {
            level.layer: level.zmin for level in self.layers.values() if level.thickness
        }

    def get_layer_to_material(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.material
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_sidewall_angle(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.sidewall_angle
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_info(self) -> Dict[Tuple[int, int], Dict]:
        """Returns layer tuple to info dict."""
        return {level.layer: level.info for level in self.layers.values()}

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        return {level_name: dict(level) for level_name, level in self.layers.items()}

    def __getitem__(self, key) -> LayerLevel:
        """Access layer stack elements."""
        if key not in self.layers:
            layers = list(self.layers.keys())
            raise ValueError(f"{key!r} not in {layers}")

        return self.layers[key]

    def get_klayout_3d_script(
        self,
        layer_views: Optional[LayerViews] = None,
        dbu: Optional[float] = 0.001,
    ) -> str:
        """Returns script for 2.5D view in KLayout.

        You can include this information in your tech.lyt

        Args:
            layer_views: optional layer_views.
            dbu: Optional database unit. Defaults to 1nm.
        """
        unetched_layers = [
            layer_name
            for layer_name, level in self.layers.items()
            if level.layer and level.layer_type == "grow"
        ]
        etch_layers = [
            layer_name
            for layer_name, level in self.layers.items()
            if level.layer and level.layer_type == "etch"
        ]

        # remove all etched layers from the grown layers
        unetched_layers_dict = defaultdict(list)
        for layer_name in etch_layers:
            level = self.layers[layer_name]
            into = level.into or []
            for layer_name_etched in into:
                unetched_layers_dict[layer_name_etched].append(layer_name)
                if layer_name_etched in unetched_layers:
                    unetched_layers.remove(layer_name_etched)

        # define layers
        out = "\n".join(
            [
                f"{layer_name} = input({level.layer[0]}, {level.layer[1]})"
                for layer_name, level in self.layers.items()
                if level.layer
            ]
        )
        out += "\n"
        out += "\n"

        # define unetched layers
        for layer_name_etched, etching_layers in unetched_layers_dict.items():
            etching_layers = " - ".join(etching_layers)
            out += f"unetched_{layer_name_etched} = {layer_name_etched} - {etching_layers}\n"

        out += "\n"

        # define slabs
        for layer_name, level in self.layers.items():
            if level.layer_type == "etch":
                into = level.into or []
                for i, layer1 in enumerate(into):
                    out += f"slab_{layer1}_{layer_name}_{i} = {layer1} &amp; {layer_name}\n"

        out += "\n"

        for layer_name, level in self.layers.items():
            layer = level.layer
            zmin = level.zmin
            zmax = zmin + level.thickness
            if dbu:
                rnd_pl = len(str(dbu).split(".")[-1])
                zmin = round(zmin, rnd_pl)
                zmax = round(zmax, rnd_pl)

            if layer is None:
                continue

            elif level.layer_type == "etch":
                name = f"{layer_name}: {level.material}"

                for i, layer1 in enumerate(into):
                    unetched_level = self.layers[layer1]
                    unetched_zmin = unetched_level.zmin
                    unetched_zmax = unetched_zmin + unetched_level.thickness

                    # slab
                    slab_layer_name = f"slab_{layer1}_{layer_name}_{i}"
                    slab_zmin = unetched_level.zmin
                    slab_zmax = unetched_zmax - level.thickness
                    name = f"{slab_layer_name}: {level.material} {layer[0]}/{layer[1]}"
                    txt = (
                        f"z("
                        f"{slab_layer_name}, "
                        f"zstart: {slab_zmin}, "
                        f"zstop: {slab_zmax}, "
                        f"name: '{name}'"
                    )
                    if layer_views:
                        txt += ", "
                        props = layer_views.get_from_tuple(layer)
                        if props.color.fill == props.color.frame:
                            txt += f"color: {props.color.fill}"
                        else:
                            txt += (
                                f"fill: {props.color.fill}, "
                                f"frame: {props.color.frame}"
                            )
                    txt += ")"
                    out += f"{txt}\n"

            elif layer_name in unetched_layers:
                name = f"{layer_name}: {level.material} {layer[0]}/{layer[1]}"

                txt = (
                    f"z("
                    f"{layer_name}, "
                    f"zstart: {zmin}, "
                    f"zstop: {zmax}, "
                    f"name: '{name}'"
                )
                if layer_views:
                    txt += ", "
                    props = layer_views.get_from_tuple(layer)
                    if props.color.fill == props.color.frame:
                        txt += f"color: {props.color.fill}"
                    else:
                        txt += (
                            f"fill: {props.color.fill}, " f"frame: {props.color.frame}"
                        )

                txt += ")"
                out += f"{txt}\n"

        out += "\n"

        for layer_name in unetched_layers_dict:
            unetched_level = self.layers[layer_name]
            layer = unetched_level.layer

            unetched_zmin = unetched_level.zmin
            unetched_zmax = unetched_zmin + unetched_level.thickness
            name = f"{slab_layer_name}: {unetched_level.material}"

            unetched_layer_name = f"unetched_{layer_name}"
            name = f"{unetched_layer_name}: {unetched_level.material} {layer[0]}/{layer[1]}"
            txt = (
                f"z("
                f"{unetched_layer_name}, "
                f"zstart: {unetched_zmin}, "
                f"zstop: {unetched_zmax}, "
                f"name: '{name}'"
            )
            if layer_views:
                txt += ", "
                props = layer_views.get_from_tuple(layer)
                if props.color.fill == props.color.frame:
                    txt += f"color: {props.color.fill}"
                else:
                    txt += f"fill: {props.color.fill}, " f"frame: {props.color.frame}"
            txt += ")"
            out += f"{txt}\n"

        return out

    def filtered(self, layers):
        return type(self)(layers={k: self.layers[k] for k in layers})


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.generic_tech import LAYER_STACK

    component = c = gf.components.grating_coupler_elliptical_trenches()
    # component = c = gf.components.taper_strip_to_ridge_trenches()

    # script = LAYER_STACK.get_klayout_3d_script()
    # print(script)

    ls = layer_stack = LAYER_STACK
    layer_to_thickness = layer_stack.get_layer_to_thickness()

    c = layer_stack.get_component_with_derived_layers(component)
    c.show(show_ports=True)

    # import pathlib
    # filepath = pathlib.Path(
    #     "/home/jmatres/gdslib/sp/temp/write_sparameters_meep_mpi.json"
    # )
    # ls_json = filepath.read_bytes()
    # ls2 = LayerStack.parse_raw(ls_json)



import pathlib
from typing import Union


def append_file_extension(
    filename: Union[str, pathlib.Path], extension: str
) -> Union[str, pathlib.Path]:
    """Try appending extension to file."""
    # Handle whether given with '.'
    if "." not in extension:
        extension = f".{extension}"

    if isinstance(filename, str) and not filename.endswith(extension):
        filename += extension

    if isinstance(filename, pathlib.Path) and not str(filename).endswith(extension):
        filename = filename.with_suffix(extension)
    return filename



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec
from gdsfactory.components.bend_s import bend_s


@gf.cell
def mmi1x2_with_sbend(
    with_sbend: bool = True,
    s_bend: ComponentSpec = bend_s,
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    """Returns 1x2 splitter for Cband.

    https://opg.optica.org/oe/fulltext.cfm?uri=oe-21-1-1310&id=248418

    Args:
        with_sbend: add sbend.
        s_bend: S-bend spec.
        cross_section: spec.
    """

    def mmi_widths(t):
        from scipy.interpolate import interp1d

        # Note: Custom width/offset functions MUST be vectorizable--you must be able
        # to call them with an array input like my_custom_width_fun([0, 0.1, 0.2, 0.3, 0.4])
        widths = np.array(
            [0.5, 0.5, 0.6, 0.7, 0.9, 1.26, 1.4, 1.4, 1.4, 1.4, 1.31, 1.2, 1.2]
        )
        xold = np.linspace(0, 1, num=len(widths))
        xnew = np.linspace(0, 1, num=100)
        f = interp1d(xold, widths, kind="cubic")
        return f(xnew)

    c = gf.Component()

    P = gf.path.straight(length=2, npoints=100)
    xs = gf.get_cross_section(cross_section)
    xs.width = mmi_widths
    c << gf.path.extrude(P, cross_section=xs)

    # Add "stub" straight sections for ports
    input_port_ref = c << gf.components.straight(
        length=0.25, cross_section=cross_section
    )
    input_port_ref.center = (-0.125, 0)
    top_output_port_ref = c << gf.components.straight(
        length=0.25, cross_section=cross_section
    )
    top_output_port_ref.center = (2.125, 0.35)
    bottom_output_port_ref = c << gf.components.straight(
        length=0.25, cross_section=cross_section
    )
    bottom_output_port_ref.center = (2.125, -0.35)
    if with_sbend:
        sbend = gf.get_component(s_bend, cross_section=cross_section)
        top_sbend = c << sbend
        bot_sbend = c << sbend
        bot_sbend.mirror([1, 0])
        top_sbend.connect("o1", destination=top_output_port_ref.ports["o2"])
        bot_sbend.connect("o1", destination=bottom_output_port_ref.ports["o2"])
        c.add_port("o1", port=input_port_ref.ports["o1"])
        c.add_port("o2", port=top_sbend.ports["o2"])
        c.add_port("o3", port=bot_sbend.ports["o2"])

    else:
        c.add_port("o1", port=input_port_ref.ports["o1"])
        c.add_port("o2", port=top_output_port_ref.ports["o2"])
        c.add_port("o3", port=bottom_output_port_ref.ports["o2"])

    return c


if __name__ == "__main__":
    # c = mmi1x2_with_sbend(with_sbend=False)
    # c = mmi1x2_with_sbend(with_sbend=True)
    c = mmi1x2_with_sbend(
        with_sbend=True,
        cross_section=dict(cross_section="strip", settings=dict(layer=(2, 0))),
    )
    c.show(show_ports=True)



from gdsfactory.pdk import get_layer_stack
from gdsfactory.simulation.sax.build_model import Model
from gdsfactory.simulation.gmeep import write_sparameters_meep
import ray
from pathlib import Path
from gdsfactory.config import sparameters_path
from gdsfactory.read import import_gds


@ray.remote(num_cpus=1)
def remote_output_from_inputs(**kwargs):
    input_component = import_gds(kwargs["input_component_file"], read_metadata=True)
    output_vector_labels = kwargs["output_vector_labels"]
    del kwargs["input_component_file"]
    del kwargs["output_vector_labels"]

    sp = write_sparameters_meep(input_component, **kwargs)

    new_inputs = []
    output_vectors = []
    for wavelength_index, wavelength in enumerate(sp["wavelengths"]):
        new_inputs.append(wavelength)
        output_vectors.append(
            [sp[output_key][wavelength_index] for output_key in output_vector_labels]
        )

    return new_inputs, output_vectors


class MeepFDTDModel(Model):
    def __init__(self, **kwargs) -> None:
        """Generic model inferred from MEEP 2.5D FDTD calculations."""
        super().__init__(**kwargs)

        self.output_vector_labels = self.define_output_vector_labels()

        self.temp_dir = kwargs.get("tempdir") or Path(sparameters_path) / "temp_ray"
        self.temp_file_str = kwargs.get("tempfile") or "temp_ray"

        self.temp_dir.mkdir(exist_ok=True, parents=True)

        self.remote_function = remote_output_from_inputs.options(
            num_cpus=self.num_cpus_per_task,  # num_gpus=self.num_gpus_per_task
        )

        return None

    # def get_Sparameters(self, input_dict, output_vector_labels):
    #     """For FDTD, results are directly S-parameters."""
    #     sp = self.get_results(self, input_dict)
    #     output_vector = [sp[output_key] for output_key in self.output_vector_labels]
    #     output_vectors.append(output_vector)
    #     return sp

    def get_output_from_inputs(self, labels, values, remote_function):
        """Get one output vector from a set of inputs.

        How to map parameters to simulation inputs depends on the target simulator.

        Arguments:
            labels: keys of the parameters
            values: values of the parameters
            remote_function: ray remote function object to use for the simulation

        Returns:
            remote function ID for delayed execution
        """
        # Prepare this specific input vector
        input_dict = dict(zip(labels, [float(value) for value in values]))
        # Parse input vector according to parameter type
        param_dict, layerstack_param_dict, litho_param_dict = self.parse_input_dict(
            input_dict
        )
        input_component = self.perturb_geometry(
            self.trainable_component(param_dict), litho_param_dict
        )
        wavelengths = self.non_trainable_parameters["wavelength"]

        sim_settings = self.simulation_settings["sim_settings"]
        sim_settings["wavelength_start"] = wavelengths.min_value
        sim_settings["wavelength_stop"] = wavelengths.max_value
        sim_settings["wavelength_points"] = wavelengths.count()

        # We cannot serialize components, so save it as gds and import instead
        value_str = [str(value) for value in values]
        current_file = self.temp_dir / "_".join(list(value_str))
        input_component_file = current_file.with_suffix(".gds")
        input_component.write_gds_with_metadata(input_component_file)

        # Define function input given parameter values and transformed layerstack/component
        function_input = dict(
            input_component_file=input_component_file,
            cores=self.num_cpus_per_task,
            output_vector_labels=self.output_vector_labels,
        )
        function_input.update(sim_settings)
        # Assign the task to a worker
        return remote_function.remote(**function_input)


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.simulation.sax.parameter import NamedParameter, LithoParameter
    from gdsfactory.technology import LayerStack

    c = gf.components.coupler_full(
        coupling_length=0.1, dx=10.0, dy=5.0, gap=0.5, dw=0.0, cross_section="strip"
    )
    c.show(show_ports=True)

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "box",
                "clad",
            )
        }
    )

    def trainable_coupler(parameters):
        return gf.components.coupler_full(
            coupling_length=parameters["coupling_length"],
            gap=parameters["gap"],
            dw=0.0,
        )

    port_symmetries_coupler = {
        "o1@0,o1@0": ["o2@0,o2@0", "o3@0,o3@0", "o4@0,o4@0"],
        "o2@0,o1@0": ["o1@0,o2@0", "o3@0,o4@0", "o4@0,o3@0"],
        "o3@0,o1@0": ["o1@0,o3@0", "o2@0,o4@0", "o4@0,o2@0"],
        "o4@0,o1@0": ["o1@0,o4@0", "o2@0,o3@0", "o3@0,o2@0"],
    }

    sim_settings = dict(
        resolution=10,
        xmargin=1.0,
        ymargin=1.0,
        is_3d=False,
        port_source_names=["o1"],
        port_symmetries=port_symmetries_coupler,
        run=True,
        overwrite=True,
        layer_stack=filtered_layerstack,
        z=0.1,
    )

    coupler_model = MeepFDTDModel(
        trainable_component=trainable_coupler,
        layerstack=filtered_layerstack,
        simulation_settings={
            "sim_settings": sim_settings,
        },
        trainable_parameters={
            "coupling_length": NamedParameter(
                min_value=5, max_value=5, nominal_value=5, step=10
            ),
            "gap": NamedParameter(
                min_value=0.2, max_value=0.2, nominal_value=0.2, step=0.3
            ),
            "erosion_dilation": LithoParameter(
                min_value=-0.2,
                max_value=0.2,
                nominal_value=0.0,
                step=0.4,
                layername="core",
            ),
        },
        non_trainable_parameters={
            "wavelength": NamedParameter(
                min_value=1.54, max_value=1.56, nominal_value=1.55, step=0.01
            ),
        },
        num_modes=1,
        num_cpus_per_task=1,
    )

    input_vectors, output_vectors = coupler_model.get_all_inputs_outputs(type="arange")
    # interpolator = coupler_model.set_nd_nd_interp()

    # import jax.numpy as jnp

    # params = jnp.stack(
    #     jnp.broadcast_arrays(
    #         jnp.asarray([6.0, 6.0, 6.0]),
    #         jnp.asarray([1.0, 1.0, 1.0]),
    #         jnp.asarray([1.55, 1.5, 1.55]),
    #         jnp.asarray([0.22, 0.22, 0.19]),
    #     ),
    #     0,
    # )

    # print(interpolator["o1@0,o2@0"](params))



"""You can define a path with a list of points combined with a cross-section.

A path can be extruded using any CrossSection returning a Component
The CrossSection defines the layer numbers, widths and offsets

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

import hashlib
import warnings
from collections.abc import Iterable
from typing import Callable, Optional, Union

import numpy as np
from numpy import mod, pi

from gdsfactory import snap
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.component_layout import (
    _GeometryHelper,
    _parse_move,
    _reflect_points,
    _rotate_points,
)
from gdsfactory.cross_section import CrossSection, Section, Transition
from gdsfactory.port import Port
from gdsfactory.typings import (
    Coordinates,
    CrossSectionSpec,
    Float2,
    LayerSpec,
    WidthTypes,
)


def _simplify(points, tolerance):
    import shapely.geometry as sg

    ls = sg.LineString(points)
    ls_simple = ls.simplify(tolerance=tolerance)
    return np.asarray(ls_simple.coords)


class Path(_GeometryHelper):
    """Path object for smooth Paths. You can extrude a Path with a CrossSection \
            to create a Component.

    Parameters:
        path: array-like[N][2], Path, or list of Paths.
            Points or Paths to append() initially.

    """

    def __init__(self, path=None):
        """Creates an empty path."""
        self.points = np.array([[0, 0]], dtype=np.float64)
        self.start_angle = 0
        self.end_angle = 0
        self.info = {}
        if path is not None:
            # If array[N][2]
            if (
                (np.asarray(path, dtype=object).ndim == 2)
                and np.issubdtype(np.array(path).dtype, np.number)
                and (np.shape(path)[1] == 2)
            ):
                self.points = np.array(path, dtype=np.float64)
                nx1, ny1 = self.points[1] - self.points[0]
                self.start_angle = np.arctan2(ny1, nx1) / np.pi * 180
                nx2, ny2 = self.points[-1] - self.points[-2]
                self.end_angle = np.arctan2(ny2, nx2) / np.pi * 180
            elif isinstance(path, Path):
                self.points = np.array(path.points, dtype=np.float64)
                self.start_angle = path.start_angle
                self.end_angle = path.end_angle
                self.info = {}
            elif np.asarray(path, dtype=object).size > 1:
                self.append(path)
            else:
                raise ValueError(
                    "Path() the `path` argument must be either blank, a path Object, "
                    "an array-like[N][2] list of points, or a list of these"
                )

    def __len__(self):
        """Returns path points."""
        return len(self.points)

    @property
    def bbox(self):
        """Returns the bounding box of the Path."""
        bbox = [
            (np.min(self.points[:, 0]), np.min(self.points[:, 1])),
            (np.max(self.points[:, 0]), np.max(self.points[:, 1])),
        ]
        return np.array(bbox)

    def append(self, path):
        """Attach Path to the end of this Path.

        The input path automatically rotates and translates such that it continues
        smoothly from the previous segment.

        Args:
            path : Path, array-like[N][2], or list of Paths
                The input path that will be appended.
        """
        # If appending another Path, load relevant variables
        if isinstance(path, Path):
            start_angle = path.start_angle
            end_angle = path.end_angle
            points = path.points
        # If array[N][2]
        elif (
            (np.asarray(path, dtype=object).ndim == 2)
            and np.issubdtype(np.array(path).dtype, np.number)
            and (np.shape(path)[1] == 2)
        ):
            points = np.asfarray(path)
            nx1, ny1 = points[1] - points[0]
            start_angle = np.arctan2(ny1, nx1) / np.pi * 180
            nx2, ny2 = points[-1] - points[-2]
            end_angle = np.arctan2(ny2, nx2) / np.pi * 180
        # If list of Paths or arrays
        elif isinstance(path, (list, tuple)):
            for p in path:
                self.append(p)
            return self
        else:
            raise ValueError(
                "Path.append() the `path` argument must be either "
                "a Path object, an array-like[N][2] list of points, or a list of these"
            )

        # Connect beginning of new points with old points
        points = _rotate_points(points, angle=self.end_angle - start_angle)
        points += self.points[-1, :] - points[0, :]

        # Update end angle
        self.end_angle = mod(end_angle + self.end_angle - start_angle, 360)

        # Concatenate old points + new points
        self.points = np.vstack([self.points, points[1:]])

        return self

    def offset(self, offset: Union[float, Callable[..., float]] = 0):
        """Offsets Path so that it follows the Path centerline plus an offset.

        The offset can either be a fixed value, or a function
        of the form my_offset(t) where t goes from 0->1

        Args:
            offset: int or float, callable. Magnitude of the offset
        """
        if offset == 0:
            points = self.points
            start_angle = self.start_angle
            end_angle = self.end_angle
        elif callable(offset):
            # Compute lengths
            dx = np.diff(self.points[:, 0])
            dy = np.diff(self.points[:, 1])
            lengths = np.cumsum(np.sqrt((dx) ** 2 + (dy) ** 2))
            lengths = np.concatenate([[0], lengths])
            # Create list of offset points and perform offset
            points = self._centerpoint_offset_curve(
                self.points,
                offset_distance=offset(lengths / lengths[-1]),
                start_angle=self.start_angle,
                end_angle=self.end_angle,
            )
            # Numerically compute start and end angles
            tol = 1e-6
            ds = tol / lengths[-1]
            ny1 = offset(ds) - offset(0)
            start_angle = np.arctan2(-ny1, tol) / np.pi * 180 + self.start_angle
            start_angle = np.round(start_angle, decimals=6)
            ny2 = offset(1) - offset(1 - ds)
            end_angle = np.arctan2(-ny2, tol) / np.pi * 180 + self.end_angle
            end_angle = np.round(end_angle, decimals=6)
        else:  # Offset is just a number
            points = self._centerpoint_offset_curve(
                self.points,
                offset_distance=offset,
                start_angle=self.start_angle,
                end_angle=self.end_angle,
            )
            start_angle = self.start_angle
            end_angle = self.end_angle

        self.points = points
        self.start_angle = start_angle
        self.end_angle = end_angle
        return self

    def move(self, origin=(0, 0), destination=None, axis=None):
        """Moves the Path from the origin point to the destination.

        Both origin and destination can be 1x2 array-like or a Port.

        Args:
            origin : array-like[2], Port Origin point of the move.
            destination : array-like[2], Port Destination point of the move.
            axis : {'x', 'y'} Direction of move.

        """
        dx, dy = _parse_move(origin, destination, axis)
        self.points += np.array([dx, dy])

        return self

    def rotate(self, angle: float = 45, center: Optional[Float2] = (0, 0)):
        """Rotates all Polygons in the Component around the specified center point.

        If no center point specified will rotate around (0,0).

        Args:
            angle: Angle to rotate the Component in degrees.
            center: array-like[2] or None. component of the Component.
        """
        if angle == 0:
            return self
        self.points = _rotate_points(self.points, angle, center)
        if self.start_angle is not None:
            self.start_angle = mod(self.start_angle + angle, 360)
        if self.end_angle is not None:
            self.end_angle = mod(self.end_angle + angle, 360)
        return self

    def mirror(self, p1: Float2 = (0, 1), p2: Float2 = (0, 0)):
        """Mirrors the Path across the line formed between the two specified points.

        ``points`` may be input as either single points [1,2]
        or array-like[N][2], and will return in kind.

        Args:
            p1: First point of the line.
            p2: Second point of the line.
        """
        self.points = _reflect_points(self.points, p1, p2)
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        if self.start_angle is not None:
            self.start_angle = mod(2 * angle - self.start_angle, 360)
        if self.end_angle is not None:
            self.end_angle = mod(2 * angle - self.end_angle, 360)
        return self

    def _centerpoint_offset_curve(
        self, points, offset_distance, start_angle, end_angle
    ):
        """Creates a offset curve (but does not account for cusps etc)\
        by computing the centerpoint offset of the supplied x and y points."""
        new_points = np.array(points, dtype=np.float64)
        dx = np.diff(points[:, 0])
        dy = np.diff(points[:, 1])
        theta = np.arctan2(dy, dx)
        theta = np.concatenate([theta[:1], theta, theta[-1:]])
        theta_mid = (np.pi + theta[1:] + theta[:-1]) / 2  # Mean angle between segments
        dtheta_int = np.pi + theta[:-1] - theta[1:]  # Internal angle between segments
        offset_distance = offset_distance / np.sin(dtheta_int / 2)
        new_points[:, 0] -= offset_distance * np.cos(theta_mid)
        new_points[:, 1] -= offset_distance * np.sin(theta_mid)
        if start_angle is not None:
            new_points[0, :] = points[0, :] + (
                np.sin(start_angle * np.pi / 180) * offset_distance[0],
                -np.cos(start_angle * np.pi / 180) * offset_distance[0],
            )
        if end_angle is not None:
            new_points[-1, :] = points[-1, :] + (
                np.sin(end_angle * np.pi / 180) * offset_distance[-1],
                -np.cos(end_angle * np.pi / 180) * offset_distance[-1],
            )
        return new_points

    def _parametric_offset_curve(self, points, offset_distance, start_angle, end_angle):
        """Creates a parametric offset (does not account for cusps etc) \
        by using gradient of the supplied x and y points."""
        x = points[:, 0]
        y = points[:, 1]
        dxdt = np.gradient(x)
        dydt = np.gradient(y)
        if start_angle is not None:
            dxdt[0] = np.cos(start_angle * np.pi / 180)
            dydt[0] = np.sin(start_angle * np.pi / 180)
        if end_angle is not None:
            dxdt[-1] = np.cos(end_angle * np.pi / 180)
            dydt[-1] = np.sin(end_angle * np.pi / 180)
        x_offset = x + offset_distance * dydt / np.sqrt(dxdt**2 + dydt**2)
        y_offset = y - offset_distance * dxdt / np.sqrt(dydt**2 + dxdt**2)
        return np.array([x_offset, y_offset]).T

    def length(self) -> float:
        """Return cumulative length."""
        x = self.points[:, 0]
        y = self.points[:, 1]
        dx = np.diff(x)
        dy = np.diff(y)
        return np.sum(np.sqrt((dx) ** 2 + (dy) ** 2))

    def curvature(self):
        """Calculates Path curvature.

        The curvature is numerically computed so areas where the curvature
        jumps instantaneously (such as between an arc and a straight segment)
        will be slightly interpolated, and sudden changes in point density
        along the curve can cause discontinuities.

        Returns:
            s : array-like[N]
                The arc-length of the Path
            K : array-like[N]
                The curvature of the Path
        """
        x = self.points[:, 0]
        y = self.points[:, 1]
        dx = np.diff(x)
        dy = np.diff(y)
        ds = np.sqrt((dx) ** 2 + (dy) ** 2)
        s = np.cumsum(ds)
        theta = np.arctan2(dy, dx)

        # Fix discontinuities arising from np.arctan2
        dtheta = np.diff(theta)
        dtheta[np.where(dtheta > np.pi)] += -2 * np.pi
        dtheta[np.where(dtheta < -np.pi)] += 2 * np.pi
        theta = np.concatenate([[0], np.cumsum(dtheta)]) + theta[0]

        K = np.gradient(theta, s, edge_order=2)
        return s, K

    def hash_geometry(self, precision: float = 1e-4) -> str:
        """Computes an SHA1 hash of the points in the Path and the start_angle and end_angle.

        Args:
            precision: Rounding precision for the the objects in the Component.  For instance,
                a precision of 1e-2 will round a point at (0.124, 1.748) to (0.12, 1.75)

        Returns:
            str Hash result in the form of an SHA1 hex digest string.

        .. code::

            hash(
                hash(First layer information: [layer1, datatype1]),
                hash(Polygon 1 on layer 1 points: [(x1,y1),(x2,y2),(x3,y3)] ),
                hash(Polygon 2 on layer 1 points: [(x1,y1),(x2,y2),(x3,y3),(x4,y4)] ),
                hash(Polygon 3 on layer 1 points: [(x1,y1),(x2,y2),(x3,y3)] ),
                hash(Second layer information: [layer2, datatype2]),
                hash(Polygon 1 on layer 2 points: [(x1,y1),(x2,y2),(x3,y3),(x4,y4)] ),
                hash(Polygon 2 on layer 2 points: [(x1,y1),(x2,y2),(x3,y3)] ),
            )
        """
        # A random offset which fixes common rounding errors intrinsic
        # to floating point math. Example: with a precision of 0.1, the
        # floating points 7.049999 and 7.050001 round to different values
        # (7.0 and 7.1), but offset values (7.220485 and 7.220487) don't
        magic_offset = 0.17048614

        final_hash = hashlib.sha1()
        p = np.ascontiguousarray(
            (self.points / precision) + magic_offset, dtype=np.int64
        )
        final_hash.update(p)
        p = np.ascontiguousarray((self.start_angle, self.end_angle), dtype=np.float64)
        final_hash.update(p)
        return final_hash.hexdigest()

    @classmethod
    def __get_validators__(cls):
        """For pydantic."""
        yield cls._validate

    @classmethod
    def _validate(cls, v):
        """Pydantic Path validator."""
        assert isinstance(v, Path), f"TypeError, Got {type(v)}, expecting Path"
        return v

    def to_dict(self):
        return self.hash_geometry()

    def plot(self) -> None:
        """Plot path in matplotlib.

        .. plot::
            :include-source:

            import gdsfactory as gf

            p = gf.path.euler(radius=10)
            p.plot()
        """
        from gdsfactory.quickplotter import quickplot

        return quickplot(self)

    def extrude(
        self,
        cross_section: Optional[CrossSectionSpec] = None,
        layer: Optional[LayerSpec] = None,
        width: Optional[float] = None,
        widths: Optional[Float2] = None,
        simplify: Optional[float] = None,
        shear_angle_start: Optional[float] = None,
        shear_angle_end: Optional[float] = None,
    ) -> Component:
        """Returns Component by extruding a Path with a CrossSection.

        A path can be extruded using any CrossSection returning a Component
        The CrossSection defines the layer numbers, widths and offsets.

        Args:
            p: a path is a list of points (arc, straight, euler).
            cross_section: to extrude.
            layer: optional layer.
            width: optional width in um.
            widths: tuple of starting and end width for a linear taper.
            simplify: Tolerance value for the simplification algorithm.
              All points that can be removed without changing the resulting
              polygon by more than the value listed here will be removed.
            shear_angle_start: an optional angle to shear the starting face by (in degrees).
            shear_angle_end: an optional angle to shear the ending face by (in degrees).

        .. plot::
            :include-source:

            import gdsfactory as gf

            p = gf.path.euler(radius=10)
            c = p.extrude(layer=(1, 0), width=0.5)
            c.plot()
        """
        return extrude(
            p=self,
            cross_section=cross_section,
            layer=layer,
            width=width,
            widths=widths,
            simplify=simplify,
            shear_angle_start=shear_angle_start,
            shear_angle_end=shear_angle_end,
        )

    def copy(self):
        """Returns a copy of the Path."""
        p = Path()
        p.info = self.info.copy()
        p.points = np.array(self.points)
        p.start_angle = self.start_angle
        p.end_angle = self.end_angle
        return p


PathFactory = Callable[..., Path]


def _sinusoidal_transition(y1, y2):
    dy = y2 - y1

    def sine(t):
        return y1 + (1 - np.cos(np.pi * t)) / 2 * dy

    return sine


def _parabolic_transition(y1, y2):
    dy = y2 - y1

    def parabolic(t):
        return y1 + np.sqrt(t) * dy

    return parabolic


def _linear_transition(y1, y2):
    dy = y2 - y1

    def linear(t):
        return y1 + t * dy

    return linear


def transition_exponential(y1, y2, exp=0.5):
    """Returns the function for an exponential transition.

    Args:
        y1: start width in um.
        y2: end width in um.
        exp: exponent.

    """
    return lambda t: y1 + (y2 - y1) * t**exp


adiabatic_polyfit_TE1550SOI_220nm = np.array(
    [
        1.02478963e-09,
        -8.65556534e-08,
        3.32415694e-06,
        -7.68408985e-05,
        1.19282177e-03,
        -1.31366332e-02,
        1.05721429e-01,
        -6.31057637e-01,
        2.80689677e00,
        -9.26867694e00,
        2.24535191e01,
        -3.90664800e01,
        4.71899278e01,
        -3.74726005e01,
        1.77381560e01,
        -1.12666286e00,
    ]
)


def transition_adiabatic(
    w1: float,
    w2: float,
    neff_w,
    wavelength: float = 1.55,
    alpha: float = 1,
    max_length: float = 200,
    num_points_ODE: int = 2000,
):
    """Returns the points for an optimal adiabatic transition for well-guided modes.

    Args:
        w1: start width in um.
        w2: end width in um.
        neff_w: a callable that returns the effective index as a function of width
            By default, use a compact model of neff(y) for fundamental 1550 nm TE mode of 220nm-thick
            core with 3.45 index, fully clad with 1.44 index. Many coefficients are needed to capture the behaviour.
        wavelength: wavelength, in same units as widths
        alpha: parameter that scales the rate of width change
                - closer to 0 means longer and more adiabatic;
                - 1 is the intuitive limit beyond which higher order modes are excited;
                - [2] reports good performance up to 1.4 for fundamental TE in SOI (for multiple core thicknesses)
        max_length: maximum length in um.
        num_points_ODE: number of samplings points for the ODE solve.

    References:
        [1] Burns, W. K., et al. "Optical waveguide parabolic coupling horns."
            Appl. Phys. Lett., vol. 30, no. 1, 1 Jan. 1977, pp. 28-30, doi:10.1063/1.89199.
        [2] Fu, Yunfei, et al. "Efficient adiabatic silicon-on-insulator waveguide taper."
            Photonics Res., vol. 2, no. 3, 1 June 2014, pp. A41-A44, doi:10.1364/PRJ.2.000A41.
    """

    # Define ODE
    def dWdx(w, x, neff_w, wavelength, alpha):
        return alpha * wavelength / (neff_w(w) * w)

    # Parse input
    if w2 < w1:
        wmin = w2
        wmax = w1
        order = -1
    else:
        wmin = w1
        wmax = w2
        order = 1

    # Solve ODE
    x = np.linspace(0, max_length, num_points_ODE)
    from scipy.integrate import odeint

    sol = odeint(dWdx, wmin, x, args=(neff_w, wavelength, alpha))

    # Extract optimal curve
    xs = x[np.where(sol[:, 0] < wmax)]
    ws = sol[:, 0][np.where(sol[:, 0] < wmax)]

    return xs, ws[::order]


def transition(
    cross_section1: CrossSection,
    cross_section2: CrossSection,
    width_type: WidthTypes = "sine",
) -> Transition:
    """Returns a smoothly-transitioning between two CrossSections.

    Only cross-sectional elements that have the `name` (as in X.add(..., name = 'wg') )
    parameter specified in both input CrosSections will be created.
    Port names will be cloned from the input CrossSections in reverse.

    Args:
        cross_section1: First CrossSection.
        cross_section2: Second CrossSection.
        width_type: sine or linear.
          Sets the type of width transition used if any widths are different
          between the two input CrossSections.

    """
    from gdsfactory.pdk import get_layer

    X1 = cross_section1
    X2 = cross_section2

    if not X1.aliases or not X2.aliases:
        raise ValueError(
            "transition() found no named sections in one "
            "or both inputs (cross_section1/cross_section2)."
        )

    layers1 = {get_layer(section.layer) for section in X1.sections}
    layers2 = {get_layer(section.layer) for section in X2.sections}
    layers1.add(get_layer(X1.layer))
    layers2.add(get_layer(X2.layer))

    has_common_layers = bool(layers1.intersection(layers2))
    if not has_common_layers:
        raise ValueError(
            f"transition() found no common layers X1 {layers1} and X2 {layers2}"
        )

    sections = []

    sections1 = [
        X1.aliases[alias] for alias in X1.aliases.keys() if alias in X2.aliases
    ]
    sections2 = [
        X2.aliases[alias] for alias in X2.aliases.keys() if alias in X1.aliases
    ]

    if X1.cladding_layers:
        sections1 += [
            Section(width=X1.width + 2 * offset, layer=layer)
            for offset, layer in zip(X1.cladding_offsets, X2.cladding_layers)
        ]
    if X2.cladding_layers:
        sections2 += [
            Section(width=X2.width + 2 * offset, layer=layer)
            for offset, layer in zip(X2.cladding_offsets, X2.cladding_layers)
        ]

    for section1, section2 in zip(sections1, sections2):
        offset1 = section1.offset
        offset2 = section2.offset
        width1 = section1.width
        width2 = section2.width

        if callable(offset1):
            offset1 = offset1(1)
        if callable(offset2):
            offset2 = offset2(0)
        if callable(width1):
            width1 = width1(1)
        if callable(width2):
            width2 = width2(0)

        offset_fun = _sinusoidal_transition(offset1, offset2)

        if width_type == "linear":
            width_fun = _linear_transition(width1, width2)
        elif width_type == "sine":
            width_fun = _sinusoidal_transition(width1, width2)
        elif width_type == "parabolic":
            width_fun = _parabolic_transition(width1, width2)
        else:
            raise ValueError(
                f"width_type={width_type!r} must be {'sine','linear','parabolic'}"
            )

        if section1.layer != section2.layer:
            hidden = True
            layer1 = get_layer(section1.layer)
            layer2 = get_layer(section2.layer)
            layer = (layer1, layer2)
        else:
            hidden = False
            layer = get_layer(section1.layer)

        s = Section(
            width=width_fun,
            offset=offset_fun,
            layer=layer,
            port_names=(section2.port_names[0], section1.port_names[1]),
            port_types=(section2.port_types[0], section1.port_types[1]),
            name=section1.name,
            hidden=hidden,
        )
        sections.append(s)

    return Transition(
        cross_section1=X1,
        cross_section2=X2,
        width_type=width_type,
        sections=sections,
        width=max([X1.width, X2.width]),
    )


@cell
def extrude(
    p: Path,
    cross_section: Optional[CrossSectionSpec] = None,
    layer: Optional[LayerSpec] = None,
    width: Optional[float] = None,
    widths: Optional[Float2] = None,
    simplify: Optional[float] = None,
    shear_angle_start: Optional[float] = None,
    shear_angle_end: Optional[float] = None,
) -> Component:
    """Returns Component extruding a Path with a cross_section.

    A path can be extruded using any CrossSection returning a Component
    The CrossSection defines the layer numbers, widths and offsets

    Args:
        p: a path is a list of points (arc, straight, euler).
        cross_section: to extrude.
        layer: optional layer to extrude.
        width: optional width to extrude.
        widths: tuple of starting and end width.
        simplify: Tolerance value for the simplification algorithm.
          All points that can be removed without changing the resulting.
          polygon by more than the value listed here will be removed.
        shear_angle_start: an optional angle to shear the starting face by (in degrees).
        shear_angle_end: an optional angle to shear the ending face by (in degrees).

    """
    from gdsfactory.pdk import (
        get_active_pdk,
        get_cross_section,
        get_grid_size,
        get_layer,
    )

    if cross_section is None and layer is None:
        raise ValueError("CrossSection or layer needed")

    if cross_section is not None and layer is not None:
        raise ValueError("Define only CrossSection or layer")

    if layer is not None and width is None and widths is None:
        raise ValueError("Need to define layer width or widths")
    elif width:
        cross_section = CrossSection(width=width, layer=layer)

    elif widths:
        cross_section = CrossSection(
            width=_linear_transition(widths[0], widths[1]), layer=layer
        )

    xsection_points = []
    c = Component()

    x = get_cross_section(cross_section)
    sections = x.sections or []
    sections = list(sections)

    if isinstance(x, CrossSection):
        snap_to_grid_nm = int(1e3 * (x.snap_to_grid or get_grid_size()))
        sections += [
            Section(
                width=x.width,
                offset=x.offset,
                layer=get_layer(x.layer),
                port_names=x.port_names,
                port_types=x.port_types,
            )
        ]

        if x.cladding_layers and x.cladding_offsets:
            for layer, cladding_offset in zip(x.cladding_layers, x.cladding_offsets):
                width = x.width(1) if callable(x.width) else x.width
                width = max(width) if isinstance(width, Iterable) else width
                sections += [
                    Section(
                        width=width + 2 * cladding_offset,
                        offset=x.offset,
                        layer=get_layer(layer),
                    )
                ]

    for section in sections:
        width = section.width
        offset = section.offset
        layer = get_layer(section.layer)
        port_names = section.port_names
        port_types = section.port_types
        hidden = section.hidden

        if isinstance(width, (int, float)) and isinstance(offset, (int, float)):
            xsection_points.append([width, offset])
        if isinstance(layer, int):
            layer = (layer, 0)
        if (
            isinstance(layer, Iterable)
            and len(layer) == 2
            and isinstance(layer[0], int)
            and isinstance(layer[1], int)
        ):
            xsection_points.append([layer[0], layer[1]])

        # print(offset, type(offset))
        if callable(offset):
            P_offset = p.copy()
            P_offset.offset(offset)
            points = P_offset.points
            start_angle = P_offset.start_angle
            end_angle = P_offset.end_angle
            offset = 0
        else:
            points = p.points
            start_angle = p.start_angle
            end_angle = p.end_angle

        if callable(width):
            # Compute lengths
            dx = np.diff(p.points[:, 0])
            dy = np.diff(p.points[:, 1])
            lengths = np.cumsum(np.sqrt(dx**2 + dy**2))
            lengths = np.concatenate([[0], lengths])
            width = width(lengths / lengths[-1])
        dy = offset + width / 2
        # _points = _shear_face(points, dy, shear_angle_start, shear_angle_end)

        points1 = p._centerpoint_offset_curve(
            points,
            offset_distance=dy,
            start_angle=start_angle,
            end_angle=end_angle,
        )
        dy = offset - width / 2
        # _points = _shear_face(points, dy, shear_angle_start, shear_angle_end)

        points2 = p._centerpoint_offset_curve(
            points,
            offset_distance=dy,
            start_angle=start_angle,
            end_angle=end_angle,
        )
        if shear_angle_start or shear_angle_end:
            _face_angle_start = (
                start_angle + shear_angle_start - 90 if shear_angle_start else None
            )
            _face_angle_end = (
                end_angle + shear_angle_end + 90 if shear_angle_end else None
            )
            points1 = _cut_path_with_ray(
                start_point=points[0],
                start_angle=_face_angle_start,
                end_point=points[-1],
                end_angle=_face_angle_end,
                path=points1,
            )
            points2 = _cut_path_with_ray(
                start_point=points[0],
                start_angle=_face_angle_start,
                end_point=points[-1],
                end_angle=_face_angle_end,
                path=points2,
            )

        # angle = start_angle + shear_angle_start + 90
        # points2 = _cut_path_with_ray(points[0], angle, points2, start=True)
        # Simplify lines using the Ramer–Douglas–Peucker algorithm
        if isinstance(simplify, bool):
            raise ValueError("simplify argument must be a number (e.g. 1e-3) or None")

        if simplify is not None:
            points1 = _simplify(points1, tolerance=simplify)
            points2 = _simplify(points2, tolerance=simplify)

        if x.snap_to_grid:
            points = snap.snap_to_grid(points, snap_to_grid_nm)
            points1 = snap.snap_to_grid(points1, snap_to_grid_nm)
            points2 = snap.snap_to_grid(points2, snap_to_grid_nm)

        # Join points together
        points_poly = np.concatenate([points1, points2[::-1, :]])

        layers = layer if hidden else [layer, layer]
        if not hidden and p.length() > 1e-3:
            c.add_polygon(points_poly, layer=layer)

        pdk = get_active_pdk()
        warn_off_grid_ports = pdk.warn_off_grid_ports

        # Add port_names if they were specified
        if port_names[0] is not None:
            port_width = width if np.isscalar(width) else width[0]
            port_orientation = (p.start_angle + 180) % 360
            center = points[0]
            face = [points1[0], points2[0]]
            face = [_rotated_delta(point, center, port_orientation) for point in face]

            if warn_off_grid_ports:
                center_snap = snap.snap_to_grid(center, snap_to_grid_nm)
                if center[0] != center_snap[0] or center[1] != center_snap[1]:
                    warnings.warn(f"Port center {center} has off-grid ports")

            port1 = c.add_port(
                port=Port(
                    name=port_names[0],
                    layer=get_layer(layers[0]),
                    port_type=port_types[0],
                    width=port_width,
                    orientation=port_orientation,
                    center=center,
                    cross_section=x.cross_section1
                    if hasattr(x, "cross_section1")
                    else x,
                    shear_angle=shear_angle_start,
                )
            )
            port1.info["face"] = face
        if port_names[1] is not None:
            port_width = width if np.isscalar(width) else width[-1]
            port_orientation = (p.end_angle) % 360
            center = points[-1]
            face = [points1[-1], points2[-1]]
            face = [_rotated_delta(point, center, port_orientation) for point in face]

            if warn_off_grid_ports:
                center_snap = snap.snap_to_grid(center, snap_to_grid_nm)

                if center[0] != center_snap[0] or center[1] != center_snap[1]:
                    warnings.warn(f"Port center {center} has off-grid ports")
            port2 = c.add_port(
                port=Port(
                    name=port_names[1],
                    layer=get_layer(layers[1]),
                    port_type=port_types[1],
                    width=port_width,
                    center=center,
                    orientation=port_orientation,
                    cross_section=x.cross_section2
                    if hasattr(x, "cross_section2")
                    else x,
                    shear_angle=shear_angle_end,
                )
            )
            port2.info["face"] = face

    c.info["length"] = float(np.round(p.length(), 3))

    if isinstance(x, CrossSection):
        if x.add_bbox:
            c = x.add_bbox(c)
        if x.add_pins:
            c = x.add_pins(c)
        if x.decorator:
            c = x.decorator(c) or c
    return c


def _rotated_delta(
    point: np.ndarray, center: np.ndarray, orientation: float
) -> np.ndarray:
    """Gets the rotated distance of a point from a center.

    Args:
        point: the initial point.
        center: a center point to use as a reference.
        orientation: the rotation, in degrees.

    Returns: the normalized delta between the point and center, accounting for rotation
    """
    ca = np.cos(orientation * np.pi / 180)
    sa = np.sin(orientation * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])
    delta = point - center
    return np.dot(delta, rot_mat)


def _cut_path_with_ray(
    start_point: np.ndarray,
    start_angle: Optional[float],
    end_point: np.ndarray,
    end_angle: Optional[float],
    path: np.ndarray,
) -> np.ndarray:
    """Cuts or extends a path given a point and angle to project."""
    import shapely.geometry as sg
    import shapely.ops

    # a distance to approximate infinity to find ray-segment intersections
    far_distance = 10000

    path_cmp = np.copy(path)
    # pad start
    dp = path[0] - path[1]
    d_ext = far_distance / np.sqrt(np.sum(dp**2)) * np.array([dp[0], dp[1]])
    path_cmp[0] += d_ext
    # pad end
    dp = path[-1] - path[-2]
    d_ext = far_distance / np.sqrt(np.sum(dp**2)) * np.array([dp[0], dp[1]])
    path_cmp[-1] += d_ext

    intersections = [sg.Point(path[0]), sg.Point(path[-1])]
    distances = []
    ls = sg.LineString(path_cmp)
    for i, angle, point in [(0, start_angle, start_point), (1, end_angle, end_point)]:
        if angle:
            # get intersection
            angle_rad = np.deg2rad(angle)
            dx_far = np.cos(angle_rad) * far_distance
            dy_far = np.sin(angle_rad) * far_distance
            d_far = np.array([dx_far, dy_far])
            ls_ray = sg.LineString([point - d_far, point + d_far])
            intersection = ls.intersection(ls_ray)

            if not isinstance(intersection, sg.Point):
                if not isinstance(intersection, sg.MultiPoint):
                    raise ValueError(
                        f"Expected intersection to be a point, but got {intersection}"
                    )
                _, nearest = shapely.ops.nearest_points(sg.Point(point), intersection)
                intersection = nearest
            intersections[i] = intersection
        else:
            intersection = intersections[i]
        distance = ls.project(intersection)
        distances.append(distance)
    # when trimming the start, start counting at the intersection point, then
    # add all subsequent points
    points = [np.array(intersections[0].coords[0])]
    for point in path[1:-1]:
        if distances[0] < ls.project(sg.Point(point)) < distances[1]:
            points.append(point)
    points.append(np.array(intersections[1].coords[0]))
    return np.array(points)


def arc(
    radius: float = 10.0,
    angle: float = 90,
    npoints: Optional[int] = None,
    start_angle: Optional[float] = -90,
) -> Path:
    """Returns a radial arc.

    Args:
        radius: minimum radius of curvature.
        angle: total angle of the curve.
        npoints: Number of points used per 360 degrees. Defaults to pdk.bend_points_distance.
        start_angle: initial angle of the curve for drawing, default -90 degrees.

    .. plot::
        :include-source:

        import gdsfactory as gf

        p = gf.path.arc(radius=10, angle=45)
        p.plot()

    """
    from gdsfactory.pdk import get_active_pdk

    PDK = get_active_pdk()

    npoints = npoints or abs(int(angle / 360 * radius / PDK.bend_points_distance / 2))
    npoints = max(npoints, int(360 / angle) + 1)

    t = np.linspace(
        start_angle * np.pi / 180, (angle + start_angle) * np.pi / 180, npoints
    )
    x = radius * np.cos(t)
    y = radius * (np.sin(t) + 1)
    points = np.array((x, y)).T * np.sign(angle)

    P = Path()
    # Manually add points & adjust start and end angles
    P.points = points
    P.start_angle = start_angle + 90
    P.end_angle = start_angle + angle + 90
    return P


def _cumtrapz(x):
    """Numpy-based implementation of the cumulative trapezoidal integration \
    function usually found in scipy (scipy.integrate.cumtrapz)."""
    return np.cumsum((x[1:] + x[:-1]) / 2)


def _fresnel(R0, s, num_pts, n_iter=8):
    """Fresnel integral using a series expansion."""
    t = np.linspace(0, s / (np.sqrt(2) * R0), num_pts)
    x = np.zeros(num_pts)
    y = np.zeros(num_pts)

    for n in range(n_iter):
        x += (-1) ** n * t ** (4 * n + 1) / (np.math.factorial(2 * n) * (4 * n + 1))
        y += (-1) ** n * t ** (4 * n + 3) / (np.math.factorial(2 * n + 1) * (4 * n + 3))

    return np.array([np.sqrt(2) * R0 * x, np.sqrt(2) * R0 * y])


def euler(
    radius: float = 10,
    angle: float = 90,
    p: float = 0.5,
    use_eff: bool = False,
    npoints: Optional[int] = None,
) -> Path:
    """Returns an euler bend that adiabatically transitions from straight to curved.

    `radius` is the minimum radius of curvature of the bend.
    However, if `use_eff` is set to True, `radius` corresponds to the effective
    radius of curvature (making the curve a drop-in replacement for an arc).
    If p < 1.0, will create a "partial euler" curve as described in Vogelbacher et. al.
    https://dx.doi.org/10.1364/oe.27.031394

    Args:
        radius: minimum radius of curvature.
        angle: total angle of the curve.
        p: Proportion of the curve that is an Euler curve.
        use_eff: If False: `radius` is the minimum radius of curvature of the bend.
            If True: The curve will be scaled such that the endpoints match an arc.
            with parameters `radius` and `angle`.
        npoints: Number of points used per 360 degrees.

    .. plot::
        :include-source:

        import gdsfactory as gf

        p = gf.path.euler(radius=10, angle=45, p=1, use_eff=True, npoints=720)
        p.plot()

    """
    from gdsfactory.pdk import get_active_pdk

    if (p < 0) or (p > 1):
        raise ValueError("euler requires argument `p` be between 0 and 1")
    if p == 0:
        P = arc(radius=radius, angle=angle, npoints=npoints)
        P.info["Reff"] = radius
        P.info["Rmin"] = radius
        return P

    if angle < 0:
        mirror = True
        angle = np.abs(angle)
    else:
        mirror = False

    R0 = 1
    alpha = np.radians(angle)
    Rp = R0 / (np.sqrt(p * alpha))
    sp = R0 * np.sqrt(p * alpha)
    s0 = 2 * sp + Rp * alpha * (1 - p)

    PDK = get_active_pdk()
    npoints = npoints or abs(int(angle / 360 * radius / PDK.bend_points_distance / 2))
    npoints = max(npoints, int(360 / angle) + 1)

    num_pts_euler = int(np.round(sp / (s0 / 2) * npoints))
    num_pts_arc = npoints - num_pts_euler

    # Ensure a minimum of 2 points for each euler/arc section
    if npoints <= 2:
        num_pts_euler = 0
        num_pts_arc = 2

    if num_pts_euler > 0:
        xbend1, ybend1 = _fresnel(R0, sp, num_pts_euler)
        xp, yp = xbend1[-1], ybend1[-1]
        dx = xp - Rp * np.sin(p * alpha / 2)
        dy = yp - Rp * (1 - np.cos(p * alpha / 2))
    else:
        xbend1 = ybend1 = np.asfarray([])
        dx = 0
        dy = 0

    s = np.linspace(sp, s0 / 2, num_pts_arc)
    xbend2 = Rp * np.sin((s - sp) / Rp + p * alpha / 2) + dx
    ybend2 = Rp * (1 - np.cos((s - sp) / Rp + p * alpha / 2)) + dy

    x = np.concatenate([xbend1, xbend2[1:]])
    y = np.concatenate([ybend1, ybend2[1:]])
    points1 = np.array([x, y]).T
    points2 = np.flipud(np.array([x, -y]).T)

    points2 = _rotate_points(points2, angle - 180)
    points2 += -points2[0, :] + points1[-1, :]

    points = np.concatenate([points1[:-1], points2])

    # Find y-axis intersection point to compute Reff
    start_angle = 180 * (angle < 0)
    end_angle = start_angle + angle
    dy = np.tan(np.radians(end_angle - 90)) * points[-1][0]
    Reff = points[-1][1] - dy
    Rmin = Rp

    # Fix degenerate condition at angle == 180
    if np.abs(180 - angle) < 1e-3:
        Reff = points[-1][1] / 2

    # Scale curve to either match Reff or Rmin
    scale = radius / Reff if use_eff else radius / Rmin
    points *= scale

    P = Path()

    # Manually add points & adjust start and end angles
    P.points = points
    P.start_angle = start_angle
    P.end_angle = end_angle
    P.info["Reff"] = Reff * scale
    P.info["Rmin"] = Rmin * scale
    if mirror:
        P.mirror((1, 0))
    return P


def straight(length: float = 10.0, npoints: int = 2) -> Path:
    """Returns a straight path.

    For transitions you should increase have at least 100 points

    Args:
        length: of straight.
        npoints: number of points.

    """
    if length < 0:
        raise ValueError(f"length = {length} needs to be > 0")
    x = np.linspace(0, length, npoints)
    y = x * 0
    points = np.array((x, y)).T

    p = Path()
    p.append(points)
    return p


def spiral_archimedean(
    min_bend_radius: float, separation: float, number_of_loops: float, npoints: int
) -> Path:
    """Returns an Archimedean spiral.

    Args:
        radius: Inner radius of the spiral.
        separation: Separation between the loops in um.
        number_of_loops: number of loops.
        npoints: number of Points.

    .. plot::
        :include-source:

        import gdsfactory as gf

        p = gf.path.spiral_archimedean(min_bend_radius=5, separation=2, number_of_loops=3, npoints=200)
        p.plot()

    """
    return Path(
        [
            (separation / np.pi * theta + min_bend_radius)
            * np.array((np.sin(theta), np.cos(theta)))
            for theta in np.linspace(0, number_of_loops * 2 * np.pi, npoints)
        ]
    )


def _compute_segments(points):
    points = np.asfarray(points)
    normals = np.diff(points, axis=0)
    normals = (normals.T / np.linalg.norm(normals, axis=1)).T
    dx = np.diff(points[:, 0])
    dy = np.diff(points[:, 1])
    ds = np.sqrt(dx**2 + dy**2)
    theta = np.degrees(np.arctan2(dy, dx))
    dtheta = np.diff(theta)
    dtheta = dtheta - 360 * np.floor((dtheta + 180) / 360)
    return points, normals, ds, theta, dtheta


def smooth(
    points: Coordinates,
    radius: float = 4.0,
    bend: PathFactory = euler,
    **kwargs,
) -> Path:
    """Returns a smooth Path from a series of waypoints.

    Args:
        points: array-like[N][2] List of waypoints for the path to follow.
        radius: radius of curvature, passed to `bend`.
        bend: bend function that returns a path that round corners.
        npoints: Number of points used per 360 degrees for the bend.
        kwargs: Extra keyword arguments that will be passed to `bend`.

    .. plot::
        :include-source:

        import gdsfactory as gf

        p = gf.path.smooth(([0, 0], [0, 10], [10, 10]))
        p.plot()

    """
    if isinstance(points, Path):
        points = points.points

    points, normals, ds, theta, dtheta = _compute_segments(points)
    colinear_elements = np.concatenate([[False], np.abs(dtheta) < 1e-6, [False]])
    if np.any(colinear_elements):
        new_points = points[~colinear_elements, :]
        points, normals, ds, theta, dtheta = _compute_segments(new_points)

    if np.any(np.abs(np.abs(dtheta) - 180) < 1e-6):
        raise ValueError(
            "smooth() received points which double-back on themselves"
            "--turns cannot be computed when going forwards then exactly backwards."
        )

    # FIXME add caching
    # Create arcs
    paths = []
    radii = []
    for dt in dtheta:
        P = bend(radius=radius, angle=dt, **kwargs)
        chord = np.linalg.norm(P.points[-1, :] - P.points[0, :])
        r = (chord / 2) / np.sin(np.radians(dt / 2))
        r = np.abs(r)
        radii.append(r)
        paths.append(P)

    d = np.abs(np.array(radii) / np.tan(np.radians(180 - dtheta) / 2))
    encroachment = np.concatenate([[0], d]) + np.concatenate([d, [0]])
    if np.any(encroachment > ds):
        raise ValueError(
            "smooth(): Not enough distance between points to to fit curves."
            "Try reducing the radius or spacing the points out farther"
        )
    p1 = points[1:-1, :] - normals[:-1, :] * d[:, np.newaxis]

    # Move arcs into position
    new_points = []
    new_points.append([points[0, :]])
    for n in range(len(dtheta)):
        P = paths[n]
        P.rotate(theta[n] - 0)
        P.move(p1[n])
        new_points.append(P.points)
    new_points.append([points[-1, :]])
    new_points = np.concatenate(new_points)

    P = Path()
    P.rotate(theta[0])
    P.append(new_points)
    P.move(points[0, :])
    return P


__all__ = [
    "straight",
    "euler",
    "arc",
    "extrude",
    "transition",
    "smooth",
    "Path",
]


def _demo() -> None:
    import gdsfactory as gf

    c = gf.Component()
    X1 = gf.CrossSection()
    X1.add(width=1.2, offset=0, layer=2, name="wg", ports=("in1", "out1"))
    X1.add(width=2.2, offset=0, layer=3, name="etch")
    X1.add(width=1.1, offset=3, layer=1, name="wg2")

    # Create the second CrossSection that we want to transition to
    X2 = gf.CrossSection()
    X2.add(width=1, offset=0, layer=2, name="wg", ports=("in2", "out2"))
    X2.add(width=3.5, offset=0, layer=3, name="etch")
    X2.add(width=3, offset=5, layer=1, name="wg2")

    Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")

    P1 = gf.path.straight(length=5)
    P2 = gf.path.straight(length=5)

    wg1 = gf.path.extrude(P1, X1)
    wg2 = gf.path.extrude(P2, X2)

    P4 = gf.path.euler(radius=25, angle=45, p=0.5, use_eff=False)
    wg_trans = gf.path.extrude(P4, Xtrans)
    wg1_ref = c << wg1
    wg2_ref = c << wg2
    wgt_ref = c << wg_trans
    wgt_ref.connect("in2", wg1_ref.ports["out1"])
    wg2_ref.connect("in2", wgt_ref.ports["out1"])

    print(wg1)
    print(wg2)
    print(wg_trans)
    c.show(show_ports=True)


def _my_custom_width_fun(t):
    # Note: Custom width/offset functions MUST be vectorizable--you must be able
    # to call them with an array input like my_custom_width_fun([0, 0.1, 0.2, 0.3, 0.4])
    num_periods = 5
    return 3 + np.cos(2 * np.pi * t * num_periods)


def _demo_variable_width() -> None:
    # Create the Path
    P = straight(length=40, npoints=40)

    # Create two cross-sections: one fixed width, one modulated by my_custom_offset_fun
    s = Section(width=_my_custom_width_fun, offset=0, layer=(1, 0))
    X = CrossSection(width=3, offset=-6, layer=(2, 0), sections=[s])

    # Extrude the Path to create the Component
    c = extrude(P, cross_section=X)
    c.show(show_ports=True)


def _my_custom_offset_fun(t):
    # Note: Custom width/offset functions MUST be vectorizable--you must be able
    # to call them with an array input like my_custom_offset_fun([0, 0.1, 0.2,
    # 0.3, 0.4])
    num_periods = 3
    return 3 + np.cos(2 * np.pi * t * num_periods)


def _demo_variable_offset() -> None:
    # Create the Path
    P = straight(length=40, npoints=30)

    # Create two cross-sections: one fixed offset, one modulated by my_custom_offset_fun
    s1 = Section(width=1, offset=_my_custom_offset_fun, layer=(2, 0))
    X = CrossSection(layer=(1, 0), width=1, sections=[s1])

    # Extrude the Path to create the Component
    c = extrude(P, cross_section=X)
    c.show(show_ports=True)


if __name__ == "__main__":
    import numpy as np

    points = np.array([(20, 10), (40, 10), (20, 40), (50, 40), (50, 20), (70, 20)])

    p = smooth(points=points)
    # p = arc(start_angle=0)
    c = p.extrude(layer=(1, 0), width=0.1)

    # p = straight()
    # p.plot()

    # from phidl.path import smooth
    # p = smooth(
    #     points=points,
    #     radius=2,
    #     # bend=gf.path.euler,
    #     use_eff=False,
    # )

    # c = p.extrude(layer=(1, 0), width=0.1)
    # c = gf.read.from_phidl(c)
    c.show()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
import gdsfactory.simulation.gtidy3d as gt

fiber_port_name = "o2"


def test_sparameters_grating_coupler(overwrite=True) -> None:
    """Checks Sparameters for a grating_coupler_elliptical_arbitrary in 2D."""
    c = gf.components.grating_coupler_elliptical_arbitrary(
        widths=[0.343] * 25, gaps=[0.345] * 25
    )

    fiber_angle_deg = 20
    offsets = [0]
    dfs = [
        gt.write_sparameters_grating_coupler(
            component=c,
            is_3d=False,
            fiber_angle_deg=fiber_angle_deg,
            fiber_xoffset=fiber_xoffset,
            overwrite=overwrite,
        )
        for fiber_xoffset in offsets
    ]
    sp = dfs[0]

    # Check reasonable reflection/transmission
    transmission = np.abs(sp[f"{fiber_port_name}@0,o1@0"])
    reflection = np.abs(sp["o1@0,o1@0"])

    assert 1 > transmission.max() > 0.2, transmission.max()
    assert 0.3 > reflection.max() > 0, reflection.max()


if __name__ == "__main__":
    overwrite = False
    c = gf.components.grating_coupler_elliptical_arbitrary(
        widths=[0.343] * 25, gaps=[0.345] * 25
    )

    fiber_angle_deg = 20
    offsets = [0, 1]
    dfs = [
        gt.write_sparameters_grating_coupler(
            component=c,
            is_3d=False,
            fiber_angle_deg=fiber_angle_deg,
            fiber_xoffset=fiber_xoffset,
            overwrite=overwrite,
        )
        for fiber_xoffset in offsets
    ]
    sp = dfs[0]
    transmission = np.abs(sp[f"{fiber_port_name}@0,o1@0"])
    reflection = np.abs(sp["o1@0,o1@0"])

    assert 1 > transmission.max() > 0.2, transmission.max()
    assert 0.3 > reflection.max() > 0, reflection.max()



# flake8: noqa
import jax
import jax.numpy as jnp
import numpy as np
import jax.random as jran


def mlp_regression(
    input_vectors,
    output_vectors,
    training_test_split: float = 0.8,
    num_layers: int = 3,
    num_neurons_per_layer: int = 5,
    learning_rate: float = 1e-2,
    num_epochs: int = 100,
    batch_size: int = 3,
    seed: int = 0,
    save: str = "weights",
):
    """Fit multilayer perceptron inference model for vectors [self.num_inputs] --> [self.num_outputs].
    Regression task using SeLu activation on all hidden layers. Adapted from https://gist.github.com/aphearin/0da99f715906715e1d7d6b004c2dbb73.
    TODO:
        (1) more flexibility to parametrize the training (adaptive learning rate, optimizer choice, etc.),
        (2) more flexibility in network topologies, activations
        (3) network hyperparameter tuning
        (4) try low-data online learning
    Arguments:
        input_vectors: vector of inputs
        output_vectors: vector of outputs (examples)
        training_test_split: % training vs test data
        num_layers: number of layers in the neural net.
        num_neurons_per_layer: number of neurons per layer.
        learning_rate: for training.
        num_epochs: for training.
        batch_size: for training.
        seed: for weight initialization
        save: where to save the model weights.
    """

    # Split into training and test data
    x = np.random.rand(jnp.shape(input_vectors)[0])
    indices = np.random.permutation(x.shape[0])
    training_idx, test_idx = (
        indices[: int(training_test_split * len(x))],
        indices[int(training_test_split * len(x)) :],
    )
    training_inputs, test_inputs = input_vectors[training_idx], input_vectors[test_idx]
    training_outputs, test_outputs = (
        output_vectors[training_idx],
        output_vectors[test_idx],
    )

    def feedforward_prediction(params, abscissa):
        """Each neuron is just the activation function applied to y = w*x + b, except for the final layer, when no activation function is used.
        Parameters
        ----------
        params : list
            Parameters of the network, with one list element per layer.
            See notes below on network initialization.
        abscissa : ndarray
            Array of shape (batch_size, n_features)
        Returns
        -------
        preds : ndarray
            Array of shape (batch_size, n_targets)
        """
        activations = abscissa

        #  Loop over every dense layer except the last
        for w, b in params[:-1]:
            outputs = jnp.dot(w, activations) + b  # apply affine transformation
            activations = jax.nn.selu(outputs)  #  apply nonlinear activation

        #  Now for the final layer
        w_final, b_final = params[-1]
        final_outputs = jnp.dot(w_final, activations) + b_final
        return final_outputs  # Final layer is just w*x + b with no activation

    def get_random_layer_params(m, n, ran_key, scale=0.01):
        """Helper function to randomly initialize.
        weights and biases using the JAX-defined randoms.
        """
        w_key, b_key = jran.split(ran_key)
        ran_weights = scale * jran.normal(w_key, (n, m))
        ran_biases = scale * jran.normal(b_key, (n,))
        return ran_weights, ran_biases

    def get_init_network_params(sizes, ran_key):
        """Initialize all layers for a fully-connected neural network."""
        keys = jran.split(ran_key, len(sizes))
        return [
            get_random_layer_params(m, n, k)
            for m, n, k in zip(sizes[:-1], sizes[1:], keys)
        ]

    def get_network_layer_sizes(n_features, n_targets, n_layers, n_neurons_per_layer):
        dense_layer_sizes = [n_neurons_per_layer] * n_layers
        layer_sizes = [n_features, *dense_layer_sizes, n_targets]
        return layer_sizes

    ran_key = jran.PRNGKey(seed)

    num_features, num_targets = (
        jnp.shape(training_inputs)[1],
        jnp.shape(training_outputs)[1],
    )

    layer_sizes = get_network_layer_sizes(
        num_features, num_targets, num_layers, num_neurons_per_layer
    )

    init_params = get_init_network_params(layer_sizes, ran_key)

    batched_prediction = jax.vmap(feedforward_prediction, in_axes=(None, 0))

    # @jax.jit
    def mse_loss(params, abscissa, targets):
        preds = batched_prediction(params, abscissa)
        diff = preds - targets
        return jnp.sum(diff * diff) / preds.shape[0]

    # @jax.jit
    def update(params, x, y, learning_rate):
        grads = jax.grad(mse_loss)(params, x, y)
        return [
            (w - learning_rate * dw, b - learning_rate * db)
            for (w, b), (dw, db) in zip(params, grads)
        ]

    params = init_params
    losses = []
    for epoch in range(num_epochs):  # noqa: B007
        # Sample batch size test and training
        training_indices = np.random.choice(
            range(len(training_inputs)), batch_size, replace=False
        )
        testing_indices = np.random.choice(
            range(len(test_inputs)), batch_size, replace=False
        )
        current_loss = mse_loss(
            params, test_inputs[testing_indices], test_outputs[testing_indices]
        )
        losses.append(current_loss)
        params = update(
            params,
            training_inputs[training_indices],
            training_outputs[training_indices],
            learning_rate,
        )

    # Format into per-mode callable
    return [
        lambda input_vector: feedforward_prediction(params, input_vector)[i]
        for i in range(num_targets)
    ]



"""Compute group and effective index for different waveguide widths and heights.

Reproduce Yufei thesis results with MPB.

https://www.photonics.intec.ugent.be/contact/people.asp?ID=332

"""

from __future__ import annotations

import pathlib

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pydantic
from scipy.interpolate import interp2d

from gdsfactory.config import PATH
from gdsfactory.simulation.modes.find_mode_dispersion import find_mode_dispersion

PATH.modes = pathlib.Path.cwd() / "data"

nm = 1e-3
width0 = 465 * nm
thickness0 = 215 * nm


@pydantic.validate_arguments
def find_neff_ng_dw_dh(
    width: float = width0,
    thickness: float = thickness0,
    delta_width: float = 30 * nm,
    delta_thickness: float = 20 * nm,
    wavelength: float = 1.55,
    steps: int = 11,
    mode_number: int = 1,
    core: str = "Si",
    clad: str = "SiO2",
    **kwargs,
) -> pd.DataFrame:
    """Computes group and effective index for different widths and heights.

    Args:
        width: nominal waveguide width in um.
        thickness: nominal waveguide thickness in um.
        delta_width: delta width max in um.
        delta_thickness: delta thickness max in um.
        wavelength: center wavelength (um).
        steps: number of steps to sweep in width and thickness.
        mode_number: mode index to compute (1: fundamental mode).
        core: core material name.
        clad: clad material name.

    Keyword Args:
        wg_thickness: wg height (um).
        sx: supercell width (um).
        sy: supercell height (um).
        resolution: (pixels/um).
        wavelength: wavelength in um.
        num_bands: mode order.
        plot: if True plots mode.
        logscale: plots in logscale.
        plotH: plot magnetic field.
        cache: path to save the modes.
        polarization: prefix when saving the modes.
        parity: symmetries mp.ODD_Y mp.EVEN_X for TE, mp.EVEN_Y for TM.

    """
    dw = np.linspace(-delta_width, delta_width, steps)
    dh = np.linspace(-delta_thickness, delta_thickness, steps)

    neffs = []
    ngs = []
    dhs = []
    dws = []

    for dwi in dw:
        for dhi in dh:
            m = find_mode_dispersion(
                core=core,
                clad=clad,
                wg_width=width + dwi,
                wg_thickness=thickness + dhi,
                wavelength=wavelength,
                mode_number=mode_number,
                **kwargs,
            )
            neffs.append(m.neff)
            ngs.append(m.ng)
            dws.append(dwi)
            dhs.append(dhi)

    return pd.DataFrame(dict(dw=dws, dh=dhs, neff=neffs, ng=ngs))


def plot_neff_ng_dw_dh(
    width: float = width0,
    thickness: float = thickness0,
    wavelength: float = 1.55,
    mode_number: int = 1,
    **kwargs,
) -> None:
    """Plot neff and group index versus width (dw) and height (dh) variations.

    Args:
        width: waveguide width in um.
        thickness: waveguide thickness in um.
        wavelength: in um.
        mode_number: 1 is the fundamental first order mode.

    """
    filepath = pathlib.Path(PATH.modes / "mpb_dw_dh_dispersion.csv")
    m = find_mode_dispersion(
        wg_width=width, wg_thickness=thickness, wavelength=wavelength
    )
    neff0 = m.neff
    ng0 = m.ng

    if filepath.exists():
        df = pd.read_csv(filepath)
    else:
        df = find_neff_ng_dw_dh(wavelength=wavelength, **kwargs)
        cache = filepath.parent
        cache.mkdir(exist_ok=True, parents=True)
        df.to_csv(filepath)

    dws = df.dw.values
    dhs = df.dh.values
    ngs = df.ng.values
    neffs = df.neff.values

    # neff interpolation
    f_w = interp2d(neffs, ngs, np.array(dws), kind="cubic")
    f_h = interp2d(neffs, ngs, np.array(dhs), kind="cubic")

    ws = width + np.array(dws)
    hs = thickness + np.array(dhs)

    plt.plot(ws * 1e3, hs * 1e3, "ko")
    extracted_dw = []
    extracted_dh = []

    for neff, ng in zip(neffs, ngs):
        temp_w = f_w(neff, ng) + width
        temp_h = f_h(neff, ng) + thickness
        extracted_dw.append(temp_w * 1e3)
        extracted_dh.append(temp_h * 1e3)

    plt.plot(extracted_dw, extracted_dh, "rx")
    plt.xlabel("width (nm)")
    plt.ylabel("height (nm)")

    plt.figure()
    plt.plot(neffs, ngs, "ro")
    plt.plot(neff0, ng0, "bx")
    plt.xlabel("neff")
    plt.ylabel("ng")
    plt.show()


if __name__ == "__main__":
    plot_neff_ng_dw_dh()



from __future__ import annotations

import contextlib
import pathlib
from collections import OrderedDict
from typing import Optional, Tuple

import numpy as np
from omegaconf import DictConfig

from gdsfactory import Component
from gdsfactory.config import PATH

c = 2.9979e8
pi = np.pi
um = 1e-6


def install_design_kit(
    session: object,
    install_dir: pathlib.Path = PATH.interconnect,
    overwrite: bool = False,
):
    from gdsfactory.pdk import get_interconnect_cml_path

    cml_path = get_interconnect_cml_path()
    session.installdesignkit(str(cml_path), str(install_dir), overwrite)


def set_named_settings(session: object, simulation_settings: dict, element: str):
    for param, val in zip(simulation_settings.keys(), simulation_settings.values()):
        session.setnamed(element, param, val)


# TODO: Add custom s-parameter models using compound elements


def add_interconnect_element(
    session: object,
    label: str,
    model: str,
    loc: Tuple[float, float] = (200.0, 200.0),
    flip_vert: bool = False,
    flip_horiz: bool = False,
    rotation: float = 0.0,
    simulation_props: Optional[OrderedDict] = None,
):
    """Add an element to the Interconnect session.

    TODO: Need to connect this to generated s-parameters and add them to the model as well

    Args:
        session: Interconnect session.
        label: label for Interconnect component.
        model:
        loc:
        flip_vert:
        flip_horiz:
        rotation:
        extra_props:

    """
    props = OrderedDict(
        [
            ("name", label),
            ("x position", loc[0]),
            ("y position", loc[1]),
            ("horizontal flipped", float(flip_horiz)),
            ("vertical flipped", float(flip_vert)),
            ("rotated", rotation),
        ]
    )
    if simulation_props:
        if "library" in simulation_props.keys():
            _ = simulation_props.pop("library")
        if "properties" in simulation_props.keys():
            props.update(simulation_props["properties"])
        else:
            props.update(simulation_props)
    return session.addelement(model, properties=props)


def get_interconnect_settings(instance):
    info = instance.info.copy()
    if "interconnect" not in info.keys():
        return {}
    settings = info["interconnect"]
    if "properties" not in settings:
        settings["properties"] = []
    if "layout_model_property_pairs" in settings.keys():
        pairs = settings.pop("layout_model_property_pairs")
        for inc_name, (layout_name, scale) in pairs.items():
            settings["properties"][inc_name] = info[layout_name] * scale
    return settings


def send_to_interconnect(
    component: Component,
    session: object,
    ports_in: dict = None,
    ports_out: dict = None,
    placements: dict = None,
    simulation_settings: OrderedDict = None,
    drop_port_prefix: str = None,
    component_distance_scaling_x: float = 1,
    component_distance_scaling_y: float = 1,
    setup_mc: bool = False,
    exclude_electrical: bool = True,
    **settings,
) -> object:
    """Send netlist components to Interconnect and connect them according to netlist.

    Args:
        component: component from which to extract netlist.
        session: Interconnect session.
        placements: x,y pairs for where to place the components in the Interconnect GUI.
        simulation_settings: global settings for Interconnect simulation.
        drop_port_prefix: if components are written with some prefix, drop up to and including
            the prefix character.  (i.e. "c1_input" -> "input").
        component_distance_scaling: scaling factor for component distances when
            laying out Interconnect schematic.

    """
    if not session:
        import lumapi

        session = lumapi.INTERCONNECT()

    install_design_kit(session=session)

    session.switchtolayout()
    session.deleteall()

    # Make compound element for circuit
    compound_element = session.addelement(
        "Compound Element", properties=OrderedDict([("name", str(component.name))])
    )
    if setup_mc:
        MC_param_element = f"::Root Element::{component.name}"
        # Add Monte-Carlo params
        session.addproperty(
            MC_param_element, "MC_uniformity_thickness", "wafer", "Matrix"
        )
        session.addproperty(MC_param_element, "MC_uniformity_width", "wafer", "Matrix")
        session.addproperty(MC_param_element, "MC_non_uniform", "wafer", "Number")
        session.addproperty(MC_param_element, "MC_grid", "wafer", "Number")
        session.addproperty(MC_param_element, "MC_resolution_x", "wafer", "Number")
        session.addproperty(MC_param_element, "MC_resolution_y", "wafer", "Number")

    # Switch groupscope to compound element so that anything added will go into it
    session.groupscope(compound_element.name)

    c = component

    netlist = c.get_netlist()

    instances: DictConfig = netlist["instances"]
    connections: DictConfig = netlist["connections"]
    placements: DictConfig = placements or netlist["placements"]
    ports: DictConfig = netlist["ports"]

    relay_count = 1
    excluded = []
    for instance in instances:
        if exclude_electrical:
            # Exclude if purely electrical
            with contextlib.suppress(Exception):
                port_types = instances[instance].full["cross_section"].port_types
                if ("electrical" in port_types) and ("optical" not in port_types):
                    excluded.append(instance)
                    continue

        loc = (
            component_distance_scaling_x * placements[instance].x,
            component_distance_scaling_y * placements[instance].y,
        )

        sim_props = get_interconnect_settings(instances[instance])

        if "model" not in sim_props.keys():
            raise KeyError(f"Please specify an interconnect model for {instance!r}")
        model = sim_props.pop("model")

        if "layout_model_port_pairs" in sim_props.keys():
            _ = sim_props.pop("layout_model_port_pairs")

        add_interconnect_element(
            session=session,
            label=instance,
            loc=loc,
            rotation=float(placements[instance].rotation),
            model=model,
            simulation_props=sim_props,
        )
        if instance in ports_in:
            # Add input port and connect to the compound element input port
            session.addport(
                f"::Root Element::{compound_element.name}",
                f"{instance}.{ports_in[instance]}",
                "input",
                "Optical Signal",
                "left",
            )
            session.connect(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "output",
                instance,
                ports_in[instance],
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "y position",
                loc[1] - 50,
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "x position",
                loc[0] - 50,
            )
            relay_count += 1
        elif instance in ports_out:
            session.addport(
                f"::Root Element::{compound_element.name}",
                f"{instance}.{ports_out[instance]}",
                "output",
                "Optical Signal",
                "right",
            )
            session.connect(
                instance,
                ports_out[instance],
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "input",
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "y position",
                loc[1] + 50,
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "x position",
                loc[0] + 50,
            )
            relay_count += 1

    if ports:
        for port in ports_in:
            input_instance, instance_port = ports[port].split(",")
            info = get_interconnect_settings(instances[input_instance])
            if "layout_model_port_pairs" in info.keys():
                instance_port = info["layout_model_port_pairs"][instance_port]
            session.addport(
                f"::Root Element::{compound_element.name}",
                str(port),
                "input",
                "Optical Signal",
                "left",
            )
            session.connect(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "output",
                input_instance,
                instance_port,
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "y position",
                session.getnamed(input_instance, "y position") - 50,
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "x position",
                session.getnamed(input_instance, "x position") - 50,
            )
            relay_count += 1
        for port in ports_out:
            output_instance, instance_port = ports[port].split(",")
            info = get_interconnect_settings(instances[output_instance])
            if "layout_model_port_pairs" in info.keys():
                instance_port = info["layout_model_port_pairs"][instance_port]
            session.addport(
                f"::Root Element::{compound_element.name}",
                str(port),
                "output",
                "Optical Signal",
                "right",
            )
            session.connect(
                output_instance,
                instance_port,
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "input",
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "y position",
                session.getnamed(output_instance, "y position") + 50,
            )
            session.setnamed(
                f"::Root Element::{compound_element.name}::RELAY_{relay_count}",
                "x position",
                session.getnamed(output_instance, "x position") + 50,
            )
            relay_count += 1

    for connection in connections:
        element2, port2 = connection.split(",")
        element1, port1 = connections[connection].split(",")
        if (element1 in excluded) or (element2 in excluded):
            continue
        if drop_port_prefix:
            # a bad way to autodetect which ports need to have prefixes dropped.
            with contextlib.suppress(Exception):
                port1 = port1[port1.index(drop_port_prefix) + 1 :]
            with contextlib.suppress(Exception):
                port2 = port2[port2.index(drop_port_prefix) + 1 :]

        # EBeam ports are not named consistently between Klayout and Interconnect..
        element1_info = get_interconnect_settings(instances[element1])
        if "layout_model_port_pairs" in element1_info.keys():
            port1 = element1_info["layout_model_port_pairs"][port1]
        element2_info = get_interconnect_settings(instances[element2])
        if "layout_model_port_pairs" in element2_info.keys():
            port2 = element2_info["layout_model_port_pairs"][port2]

        session.connect(element1, port1, element2, port2)

    if simulation_settings:
        set_named_settings(
            session, simulation_settings, element=f"::Root Element::{component.name}"
        )
    # Return to highest-level element
    session.groupscope("::Root Element")

    # Auto-distribute ports on the compound element
    session.autoarrange(component.name)
    return session


def run_wavelength_sweep(
    component: Component,
    session: Optional[object] = None,
    setup_simulation: bool = True,
    is_top_level: bool = False,
    ports_in: Optional[dict] = None,
    ports_out: Optional[dict] = None,
    mode: int = 1,
    wavelength_range: Tuple[float, float] = (1.500, 1.600),
    n_points: int = 1000,
    results: Tuple[str, ...] = ("transmission",),
    extra_ona_props: Optional[dict] = None,
    **kwargs,
) -> dict:
    """Args are the following.

    component:
    session:
    setup_simulation: whether to send the component to interconnect before running the sweep.
    ports_in: specify the port in the Interconnect model to attach the ONA output to.
    ports_out: specify the ports in the Interconnect models to attach the ONA input to.
    wavelength_range:
    n_points:
    results:
    extra_ona_props:
    kwargs:
    """
    if len(ports_in) > 1:
        raise ValueError("Only 1 input port is supported at this time")

    import lumapi

    if not session:
        session = lumapi.INTERCONNECT()

    install_design_kit(session=session)

    if setup_simulation:
        session = send_to_interconnect(
            component=component,
            session=session,
            ports_in=ports_in,
            ports_out=ports_out,
            **kwargs,
        )

    ona_props = OrderedDict(
        [
            ("number of input ports", len(ports_out)),
            ("number of points", n_points),
            ("input parameter", "start and stop"),
            ("start frequency", (c / (wavelength_range[1] * um))),
            ("stop frequency", (c / (wavelength_range[0] * um))),
            ("plot kind", "wavelength"),
            ("relative to center", float(False)),
        ]
    )
    if extra_ona_props:
        ona_props.update(extra_ona_props)

    ona = add_interconnect_element(
        session=session,
        model="Optical Network Analyzer",
        label="ONA_1",
        loc=(0, -50),
        simulation_props=ona_props,
    )
    for port in ports_in.keys():
        name = port if is_top_level else f"{port}.{ports_in[port]}"
        session.connect(ona.name, "output", component.name, name)
    for i, port in enumerate(ports_out.keys()):
        name = port if is_top_level else f"{port}.{ports_out[port]}"
        session.connect(ona.name, f"input {i+1}", component.name, name)

    session.run()

    # inc.close()
    return {
        result: {
            port: session.getresult(ona.name, f"input {i+1}/mode {mode}/{result}")
            for i, port in enumerate(ports_out)
        }
        for result in results
    }


def plot_wavelength_sweep(
    ports_out, results, result_name: str = "TE Transmission", show: bool = True
):
    import matplotlib.pyplot as plt

    for port in ports_out:
        wl = results["transmission"][port]["wavelength"] / um
        T = 10 * np.log10(np.abs(results["transmission"][port][result_name]))
        plt.plot(wl, T, label=str(port))
    plt.legend()
    plt.xlabel(r"Wavelength ($\mu$m)")
    plt.ylabel(f"{result_name} (dB)")

    if show:
        plt.show()


if __name__ == "__main__":
    import ubcpdk.components as pdk

    mzi = pdk.mzi()

    netlist = mzi.get_netlist()

    ports_in = {"o1": "o1"}
    ports_out = {"o2": "o2"}

    simulation_settings = OrderedDict(
        [
            ("MC_uniformity_thickness", np.array([200, 200])),
            ("MC_uniformity_width", np.array([200, 200])),
            ("MC_non_uniform", 0),
            ("MC_grid", 1e-5),
            ("MC_resolution_x", 200),
            ("MC_resolution_y", 0),
        ]
    )
    results = run_wavelength_sweep(
        component=mzi,
        ports_in=ports_in,
        ports_out=ports_out,
        results=("transmission",),
        component_distance_scaling=50,
        simulation_settings=simulation_settings,
        setup_mc=True,
        is_top_level=True,
    )

    plot_wavelength_sweep(ports_out=ports_out, results=results)



from typing import Callable

import numpy as np
from pydantic import BaseModel

import gdsfactory as gf
from gdsfactory.types import Layer


class DopingLayerLevel(BaseModel):
    """Level for doping layer.

    Parameters:
        layer: (GDSII Layer number, GDSII datatype).
        type: str: "Acceptor" or "Donor".
        z_profile: callable evaluating doping versus depth.
        xy_profile: callable modulating z_profile at the edge of the layer.

    """

    layer: Layer
    type: str
    z_profile: Callable
    # xy_profile: Optional[Callable] = None # not implemented yet

    class Config:
        """pydantic config."""

        frozen = True
        extra = "forbid"


cm3_to_um3 = 1e-12


def get_doping_info_generic(
    n_conc: float = 1e17,  # * cm3_to_um3,
    p_conc: float = 1e17,  # * cm3_to_um3,
    npp_conc: float = 1e18,  # * cm3_to_um3,
    ppp_conc: float = 1e18,  # * cm3_to_um3,
):
    layermap = gf.generic_tech.LayerMap()

    return {
        "N": DopingLayerLevel(
            layer=layermap.N,
            type="Donor",
            z_profile=step(n_conc),
        ),
        "P": DopingLayerLevel(
            layer=layermap.P,
            type="Acceptor",
            z_profile=step(p_conc),
        ),
        "NPP": DopingLayerLevel(
            layer=layermap.NPP,
            type="Donor",
            z_profile=step(npp_conc),
        ),
        "PPP": DopingLayerLevel(
            layer=layermap.PPP,
            type="Acceptor",
            z_profile=step(ppp_conc),
        ),
    }


# def get_doping_xyz():
#     """Returns acceptor  vs x,y,z from DopingLayerLevel and a component."""

# def get_doping_density():
#     """Returns acceptor  vs x,y,z from DopingLayerLevel and a component."""


# def get_net_doping(component: Component, doping_info: Dict):
#     """Returns net doping from DopingLayerLevel and a component."""

#     for name, dopinglevel in doping_info.items():
#         print("ok")


def step(c, zmin=-np.inf, zmax=np.inf):
    """Step function doping of value c, between zmin and zmax."""
    return lambda y: c


# def step(c, zmin=-np.inf, zmax=np.inf):
#     """Step function doping of value c, between zmin and zmax."""
#     return lambda x, y, z: np.heaviside(c - zmin) - np.heaviside(c - zmax)

# def gaussian(n0, range, straggle):
#     """Gaussian function doping of value c."""
#     return n0*np.exp(-(z - range)**2/(2*straggle**2))

# def pearsonIV(n0, range, straggle):
#     """Gaussian function doping of value c."""
#     return n0*np.exp(-(z - range)**2/(2*straggle**2))



"""Find GDS labels and write them to a CSV file."""

from __future__ import annotations

import csv
import pathlib
from pathlib import Path
from typing import Iterator, Tuple, List

import numpy as np

import gdsfactory as gf
from gdsfactory.config import logger
from gdsfactory.pdk import get_layer
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.typings import LayerSpec, Optional, PathType


def find_labels(
    gdspath: PathType, layer_label: LayerSpec = "LABEL", prefix: str = "opt_"
) -> Iterator[Tuple[str, float, float]]:
    """Return text label and locations iterator from a GDS file.

    Klayout does not support label rotations.

    Args:
        gdspath: for the gds.
        layer_label: for the labels.
        prefix: for the labels to select.

    Returns
        string: for the label.
        x: x position (um).
        y: y position (um).
        angle: in degrees.

    """
    import klayout.db as pya

    # Load the layout
    gdspath = str(gdspath)
    layout = pya.Layout()
    layout.read(gdspath)

    layer_label = get_layer(layer_label)

    # Get the top cell and the units, and find out the index of the layer
    topcell = layout.top_cell()
    dbu = layout.dbu
    layer = pya.LayerInfo(layer_label[0], layer_label[1])
    layer_index = layout.layer(layer)

    # Extract locations
    iterator = topcell.begin_shapes_rec(layer_index)

    while not (iterator.at_end()):
        shape, trans = iterator.shape(), iterator.trans()
        iterator.next()
        if shape.is_text():
            text = shape.text
            if text.string.startswith(prefix):
                transformed = text.transformed(trans)
                yield text.string, transformed.x * dbu, transformed.y * dbu, trans.angle


def write_labels_klayout(
    gdspath: PathType,
    layer_label: LayerSpec = "LABEL",
    filepath: Optional[PathType] = None,
    prefix: str = "opt_",
) -> Path:
    """Load GDS and extracts labels in KLayout text and coordinates.

    Returns CSV filepath with each row:
    - Text
    - x
    - y
    - rotation (degrees)

    Args:
        gdspath: for the mask.
        layer_label: for labels to write.
        filepath: for CSV file. Defaults to gdspath with CSV suffix.
        prefix: for the labels to write.

    """
    labels = list(find_labels(gdspath, layer_label=layer_label, prefix=prefix))
    gdspath = pathlib.Path(gdspath)

    filepath = filepath or gdspath.with_suffix(".csv")

    with open(filepath, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerows(labels)
    logger.info(f"Wrote {len(labels)} labels to CSV {filepath.absolute()}")
    return filepath


def write_labels_gdstk(
    gdspath: Path,
    prefixes: List[str] = ("opt", "elec"),
    layer_label: LayerSpec = "LABEL",
    filepath: Optional[PathType] = None,
    debug: bool = False,
) -> Path:
    """Load GDS and extracts label text and coordinates.

    Returns CSV filepath with each row:
    - Text
    - x
    - y
    - rotation (degrees)

    Args:
        gdspath: for the mask.
        prefix: for the labels to write.
        layer_label: for labels to write.
        filepath: for CSV file. Defaults to gdspath with CSV suffix.
        debug: prints the label.
    """
    gdspath = pathlib.Path(gdspath)
    filepath = filepath or gdspath.with_suffix(".csv")
    filepath = pathlib.Path(filepath)
    c = gf.import_gds(gdspath)

    labels = [("text", "x", "y", "rotation")]
    layer_label = get_layer(layer_label)

    for label in c.get_labels():
        for prefix in prefixes:
            if (
                layer_label
                and label.layer == layer_label[0]
                and label.texttype == layer_label[1]
                and label.text.startswith(prefix)
            ):
                x, y = label.origin
                rot = np.rad2deg(label.rotation)
                labels += [(label.text, x, y, rot)]
                if debug:
                    print(label.text)

    with open(filepath, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerows(labels)
    logger.info(f"Wrote {len(labels)} labels to {filepath.absolute()}")
    return filepath


def test_find_labels() -> None:
    import gdsfactory as gf

    c = gf.components.straight(length=124)
    cc = add_fiber_single(component=c, decorator=gf.add_labels.add_labels_to_ports)
    gdspath = cc.write_gds()
    assert len(list(find_labels(gdspath))) == 4


if __name__ == "__main__":
    test_find_labels()

    # import gdsfactory as gf
    # c = gf.components.straight()
    # cc = add_fiber_single(component=c)
    # gdspath = cc.write_gds()
    # print(len(list(find_labels(gdspath))))
    # cc.show(show_ports=True)
    # gdspath = CONFIG["samples_path"] / "mask" / "build" / "mask" / "sample.gds"
    # write_labels(gdspath)
